# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.

"""
Python wrapper functions for the ``C`` package:
Basic Linear Algebra Subroutines (BLAS)

See also:
http://en.wikipedia.org/wiki/Basic_Linear_Algebra_Subprograms
and
http://www.netlib.org/lapack/lug/node145.html
"""

import numpy as np

from gpaw.utilities import is_contiguous
from gpaw import debug
import _gpaw


def mmm(alpha, a, opa, b, opb, beta, c):
    """Matrix-matrix multiplication using dgemm or zgemm.

    For opa='n' and opb='n', we have::

        c <- alpha * a * b + beta * c.

    Use 't' to transpose matrices and 'c' to transpose and complex conjugate
    matrices.
    """

    assert opa in 'NTC'
    assert opb in 'NTC'

    if opa == 'N':
        a1, a2 = a.shape
    else:
        a2, a1 = a.shape
    if opb == 'N':
        b1, b2 = b.shape
    else:
        b2, b1 = b.shape
    assert a2 == b1
    assert c.shape == (a1, b2)

    assert a.strides[1] == b.strides[1] == c.strides[1] == c.itemsize
    assert a.dtype == b.dtype == c.dtype
    if a.dtype == float:
        assert not isinstance(alpha, complex)
        assert not isinstance(beta, complex)
    else:
        assert a.dtype == complex

    _gpaw.mmm(alpha, a, opa, b, opb, beta, c)


def scal(alpha, x):
    """alpha x

    Performs the operation::

      x <- alpha * x

    """
    if isinstance(alpha, complex):
        assert is_contiguous(x, complex)
    else:
        assert isinstance(alpha, float)
        assert x.dtype in [float, complex]
        assert x.flags.contiguous
    _gpaw.scal(alpha, x)


def gemm(alpha, a, b, beta, c, transa='n'):
    """General Matrix Multiply.

    Performs the operation::

      c <- alpha * b.a + beta * c

    If transa is "n", ``b.a`` denotes the matrix multiplication defined by::

                      _
                     \
      (b.a)        =  ) b  * a
           ijkl...   /_  ip   pjkl...
                      p

    If transa is "t" or "c", ``b.a`` denotes the matrix multiplication
    defined by::

                      _
                     \
      (b.a)        =  ) b    *    a
           ij        /_  iklm...   jklm...
                     klm...

    where in case of "c" also complex conjugate of a is taken.
    """
    assert np.isfinite(c).all()

    assert (a.dtype == float and b.dtype == float and c.dtype == float and
            isinstance(alpha, float) and isinstance(beta, float) or
            a.dtype == complex and b.dtype == complex and c.dtype == complex)
    if transa == 'n':
        assert a.size == 0 or a[0].flags.contiguous
        assert c.flags.contiguous or c.ndim == 2 and c.strides[1] == c.itemsize
        assert b.ndim == 2
        assert b.strides[1] == b.itemsize
        assert a.shape[0] == b.shape[1]
        assert c.shape == b.shape[0:1] + a.shape[1:]
    else:
        assert a.flags.contiguous
        assert b.size == 0 or b[0].flags.contiguous
        assert c.strides[1] == c.itemsize
        assert a.shape[1:] == b.shape[1:]
        assert c.shape == (b.shape[0], a.shape[0])
    _gpaw.gemm(alpha, a, b, beta, c, transa)


def gemv(alpha, a, x, beta, y, trans='t'):
    """General Matrix Vector product.

    Performs the operation::

      y <- alpha * a.x + beta * y

    ``a.x`` denotes matrix multiplication, where the product-sum is
    over the entire length of the vector x and
    the first dimension of a (for trans='n'), or
    the last dimension of a (for trans='t' or 'c').

    If trans='c', the complex conjugate of a is used. The default is
    trans='t', i.e. behaviour like np.dot with a 2D matrix and a vector.

    Example::

      >>> y_m = np.dot(A_mn, x_n)
      >>> # or better yet
      >>> y_m = np.zeros(A_mn.shape[0], A_mn.dtype)
      >>> gemv(1.0, A_mn, x_n, 0.0, y_m)

    """
    assert (a.dtype == float and x.dtype == float and y.dtype == float and
            isinstance(alpha, float) and isinstance(beta, float) or
            a.dtype == complex and x.dtype == complex and y.dtype == complex)
    assert a.flags.contiguous
    assert y.flags.contiguous
    assert x.ndim == 1
    assert y.ndim == a.ndim - 1
    if trans == 'n':
        assert a.shape[0] == x.shape[0]
        assert a.shape[1:] == y.shape
    else:
        assert a.shape[-1] == x.shape[0]
        assert a.shape[:-1] == y.shape
    _gpaw.gemv(alpha, a, x, beta, y, trans)


def axpy(alpha, x, y):
    """alpha x plus y.

    Performs the operation::

      y <- alpha * x + y

    """
    if isinstance(alpha, complex):
        assert is_contiguous(x, complex) and is_contiguous(y, complex)
    else:
        assert isinstance(alpha, float)
        assert x.dtype in [float, complex]
        assert x.dtype == y.dtype
        assert x.flags.contiguous and y.flags.contiguous
    assert x.shape == y.shape
    _gpaw.axpy(alpha, x, y)


def czher(alpha, x, a):
    """alpha x * x.conj() + a.

    Performs the operation::

      y <- alpha * x * x.conj() + a

    where x is a N element vector and a is a N by N hermitian matrix, alpha
    is a real scalar.
    """

    assert isinstance(alpha, float)
    assert is_contiguous(x, complex) and is_contiguous(a, complex)
    assert x.flags.contiguous and a.flags.contiguous
    assert x.ndim == 1 and a.ndim == 2
    assert x.shape[0] == a.shape[0]

    _gpaw.czher(alpha, x, a)


def rk(alpha, a, beta, c, trans='c'):
    """Rank-k update of a matrix.

    Performs the operation::

                        dag
      c <- alpha * a . a    + beta * c

    where ``a.b`` denotes the matrix multiplication defined by::

                 _
                \
      (a.b)   =  ) a         * b
           ij   /_  ipklm...     pjklm...
               pklm...

    ``dag`` denotes the hermitian conjugate (complex conjugation plus a
    swap of axis 0 and 1).

    Only the lower triangle of ``c`` will contain sensible numbers.
    """
    assert beta == 0.0 or np.isfinite(c).all()

    assert (a.dtype == float and c.dtype == float or
            a.dtype == complex and c.dtype == complex)
    assert a.flags.contiguous
    assert a.ndim > 1
    if trans == 'n':
        assert c.shape == (a.shape[1], a.shape[1])
    else:
        assert c.shape == (a.shape[0], a.shape[0])
    assert c.strides[1] == c.itemsize
    _gpaw.rk(alpha, a, beta, c, trans)


def r2k(alpha, a, b, beta, c):
    """Rank-2k update of a matrix.

    Performs the operation::

                        dag        cc       dag
      c <- alpha * a . b    + alpha  * b . a    + beta * c

    where ``a.b`` denotes the matrix multiplication defined by::

                 _
                \
      (a.b)   =  ) a         * b
           ij   /_  ipklm...     pjklm...
               pklm...

    ``cc`` denotes complex conjugation.

    ``dag`` denotes the hermitian conjugate (complex conjugation plus a
    swap of axis 0 and 1).

    Only the lower triangle of ``c`` will contain sensible numbers.
    """
    assert beta == 0.0 or np.isfinite(np.tril(c)).all()

    assert (a.dtype == float and b.dtype == float and c.dtype == float or
            a.dtype == complex and b.dtype == complex and c.dtype == complex)
    assert a.flags.contiguous and b.flags.contiguous
    assert a.ndim > 1
    assert a.shape == b.shape
    assert c.shape == (a.shape[0], a.shape[0])
    assert c.strides[1] == c.itemsize
    _gpaw.r2k(alpha, a, b, beta, c)


def dotc(a, b):
    """Dot product, conjugating the first vector with complex arguments.

    Returns the value of the operation::

        _
       \   cc
        ) a       * b
       /_  ijk...    ijk...
       ijk...

    ``cc`` denotes complex conjugation.
    """
    assert ((is_contiguous(a, float) and is_contiguous(b, float)) or
            (is_contiguous(a, complex) and is_contiguous(b, complex)))
    assert a.shape == b.shape
    return _gpaw.dotc(a, b)


def dotu(a, b):
    """Dot product, NOT conjugating the first vector with complex arguments.

    Returns the value of the operation::

        _
       \
        ) a       * b
       /_  ijk...    ijk...
       ijk...


    """
    assert ((is_contiguous(a, float) and is_contiguous(b, float)) or
            (is_contiguous(a, complex) and is_contiguous(b, complex)))
    assert a.shape == b.shape
    return _gpaw.dotu(a, b)


def _gemmdot(a, b, alpha=1.0, beta=1.0, out=None, trans='n'):
    """Matrix multiplication using gemm.

    return reference to out, where::

      out <- alpha * a . b + beta * out

    If out is None, a suitably sized zero array will be created.

    ``a.b`` denotes matrix multiplication, where the product-sum is
    over the last dimension of a, and either
    the first dimension of b (for trans='n'), or
    the last dimension of b (for trans='t' or 'c').

    If trans='c', the complex conjugate of b is used.
    """
    # Store original shapes
    ashape = a.shape
    bshape = b.shape

    # Vector-vector multiplication is handled by dotu
    if a.ndim == 1 and b.ndim == 1:
        assert out is None
        if trans == 'c':
            return alpha * _gpaw.dotc(b, a)  # dotc conjugates *first* argument
        else:
            return alpha * _gpaw.dotu(a, b)

##     # Use gemv if a or b is a vector, and the other is a matrix??
##     if a.ndim == 1 and trans == 'n':
##         gemv(alpha, b, a, beta, out, trans='n')
##     if b.ndim == 1 and trans == 'n':
##         gemv(alpha, a, b, beta, out, trans='t')

    # Map all arrays to 2D arrays
    a = a.reshape(-1, a.shape[-1])
    if trans == 'n':
        b = b.reshape(b.shape[0], -1)
        outshape = a.shape[0], b.shape[1]
    else:  # 't' or 'c'
        b = b.reshape(-1, b.shape[-1])

    # Apply BLAS gemm routine
    outshape = a.shape[0], b.shape[trans == 'n']
    if out is None:
        # (ATLAS can't handle uninitialized output array)
        out = np.zeros(outshape, a.dtype)
    else:
        out = out.reshape(outshape)
    gemm(alpha, b, a, beta, out, trans)

    # Determine actual shape of result array
    if trans == 'n':
        outshape = ashape[:-1] + bshape[1:]
    else:  # 't' or 'c'
        outshape = ashape[:-1] + bshape[:-1]
    return out.reshape(outshape)


def _rotate(in_jj, U_ij, a=1., b=0., out_ii=None, work_ij=None):
    """Perform matrix rotation using gemm

    For the 2D input matrices in, U, do the rotation::

      out <- a * U . in . U^d + b * out

    where '.' denotes matrix multiplication and '^d' the hermitian conjugate.

    work_ij is a temporary work array for storing the intermediate product.
    out_ii, and work_ij are created if not given.

    The method returns a reference to out.
    """
    if work_ij is None:
        work_ij = np.zeros_like(U_ij)
    if out_ii is None:
        out_ii = np.zeros(U_ij.shape[:1] * 2, U_ij.dtype)
    if in_jj.dtype == float:
        trans = 't'
    else:
        trans = 'c'
    gemm(1., in_jj, U_ij, 0., work_ij, 'n')
    gemm(a, U_ij, work_ij, b, out_ii, trans)
    return out_ii


if not debug:
    mmm = _gpaw.mmm  # noqa
    scal = _gpaw.scal  # noqa
    gemm = _gpaw.gemm  # noqa
    gemv = _gpaw.gemv  # noqa
    axpy = _gpaw.axpy  # noqa
    rk = _gpaw.rk  # noqa
    r2k = _gpaw.r2k  # noqa
    dotc = _gpaw.dotc  # noqa
    dotu = _gpaw.dotu  # noqa
    gemmdot = _gemmdot
    rotate = _rotate
else:
    def gemmdot(a, b, alpha=1.0, beta=1.0, out=None, trans='n'):
        assert a.flags.contiguous
        assert b.flags.contiguous
        assert a.dtype == b.dtype
        if trans == 'n':
            assert a.shape[-1] == b.shape[0]
        else:
            assert a.shape[-1] == b.shape[-1]
        if out is not None:
            assert out.flags.contiguous
            assert a.dtype == out.dtype
            assert a.ndim > 1 or b.ndim > 1
            if trans == 'n':
                assert out.shape == a.shape[:-1] + b.shape[1:]
            else:
                assert out.shape == a.shape[:-1] + b.shape[:-1]
        return _gemmdot(a, b, alpha, beta, out, trans)

    def rotate(in_jj, U_ij, a=1., b=0., out_ii=None, work_ij=None):
        assert in_jj.dtype == U_ij.dtype
        assert in_jj.flags.contiguous
        assert U_ij.flags.contiguous
        assert in_jj.shape == U_ij.shape[1:] * 2
        if out_ii is not None:
            assert out_ii.dtype == in_jj.dtype
            assert out_ii.flags.contiguous
            assert out_ii.shape == U_ij.shape[:1] * 2
        if work_ij is not None:
            assert work_ij.dtype == in_jj.dtype
            assert work_ij.flags.contiguous
            assert work_ij.shape == U_ij.shape
        return _rotate(in_jj, U_ij, a, b, out_ii, work_ij)



"""Compare TPSS from scratch and from PBE"""
from ase import Atoms
from gpaw import GPAW, MixerSum, Davidson

n = Atoms('N', magmoms=[3])
n.center(vacuum=2.5)
def getkwargs():
    return dict(eigensolver=Davidson(4), mixer=MixerSum(0.5, 5, 10.0))

n.calc = GPAW(xc='TPSS', **getkwargs())
e1 = n.get_potential_energy()

n.calc = GPAW(xc='PBE', **getkwargs())
n.get_potential_energy()
n.calc.set(xc='TPSS')
e2 = n.get_potential_energy()
err = abs(e2 - e1)
print('Energy difference', err)
assert err < 3e-5, err



from __future__ import print_function
from gpaw.utilities.blas import czher, axpy
import numpy as np
from time import time

alpha = 0.5
x = np.random.rand(3) + 1j * np.random.rand(3)
a = np.random.rand(9).reshape(3,3) + np.random.rand(9).reshape(3,3) * 1j

# make a hermitian
for i in range(3):
    for j in range(3):
        a[i,j] = a[j,i].conj()
    a[i,i] = np.real(a[i,i])

b = alpha * np.outer(x.conj(), x) + a
czher(alpha, x, a)

for i in range(3):
    for j in range(i,3):
        a[j,i] = a[i,j].conj()
        
assert np.abs(b-a).sum() < 1e-14

# testing speed
t_czher = 0
t_axpy = 0

for i in np.arange(1000):
    t0 = time()
    czher(alpha, x, a)
    t_czher += time() - t0

    t0 = time()
    xx = np.outer(x.conj(), x)
    axpy(alpha, xx, a)
    t_axpy += time() - t0

print('t_czher:', t_czher)
print('t_axpy:', t_axpy)



import numpy as np
from ase.units import Hartree

from gpaw.xc import XC
from gpaw.utilities import unpack


def vxc(paw, xc=None, coredensity=True):
    """Calculate XC-contribution to eigenvalues."""

    ham = paw.hamiltonian
    dens = paw.density
    wfs = paw.wfs

    if xc is None:
        xc = ham.xc
    elif isinstance(xc, str):
        xc = XC(xc)

    if dens.nt_sg is None:
        dens.interpolate_pseudo_density()

    thisisatest = not True

    if xc.orbital_dependent:
        paw.get_xc_difference(xc)

    # Calculate XC-potential:
    vxct_sg = ham.finegd.zeros(wfs.nspins)
    xc.calculate(dens.finegd, dens.nt_sg, vxct_sg)
    vxct_sG = ham.restrict_and_collect(vxct_sg)
    if thisisatest:
        vxct_sG[:] = 1

    # ... and PAW corrections:
    dvxc_asii = {}
    for a, D_sp in dens.D_asp.items():
        dvxc_sp = np.zeros_like(D_sp)
        xc.calculate_paw_correction(wfs.setups[a], D_sp, dvxc_sp, a=a,
                                    addcoredensity=coredensity)
        dvxc_asii[a] = [unpack(dvxc_p) for dvxc_p in dvxc_sp]
        if thisisatest:
            dvxc_asii[a] = [wfs.setups[a].dO_ii]

    vxc_un = np.empty((wfs.kd.mynks, wfs.bd.mynbands))
    for u, vxc_n in enumerate(vxc_un):
        kpt = wfs.kpt_u[u]
        vxct_G = vxct_sG[kpt.s]
        for n in range(wfs.bd.mynbands):
            psit_G = wfs._get_wave_function_array(u, n, realspace=True)
            vxc_n[n] = wfs.gd.integrate((psit_G * psit_G.conj()).real,
                                        vxct_G, global_integral=False)

        for a, dvxc_sii in dvxc_asii.items():
            P_ni = kpt.P_ani[a]
            vxc_n += (np.dot(P_ni, dvxc_sii[kpt.s]) *
                      P_ni.conj()).sum(1).real

    wfs.gd.comm.sum(vxc_un)
    vxc_skn = wfs.kd.collect(vxc_un)

    if xc.orbital_dependent:
        vxc_skn += xc.exx_skn

    return wfs.bd.collect(vxc_skn.T.copy(), broadcast=True).T * Hartree



"""Test automagical calculation of wfs"""

from gpaw import GPAW
from ase import Atoms

# H2
H = Atoms('HH', [(0, 0, 0), (0, 0, 1)])
H.center(vacuum=2.0)

calc = GPAW(nbands=2,
            convergence={'eigenstates': 1e-3})
H.set_calculator(calc)
H.get_potential_energy()
calc.write('tmp')

calc = GPAW('tmp')
calc.converge_wave_functions()

calc.set(nbands=5)
calc.converge_wave_functions()



from gpaw.poisson import NonPeriodicLauePoissonSolver, GeneralizedLauePoissonSolver, FDPoissonSolver, idst2, dst2
from gpaw.grid_descriptor import GridDescriptor
import numpy as np

gd = GridDescriptor((10,12,42), (4, 5, 20), pbc_c=(True,True,False))
poisson = GeneralizedLauePoissonSolver(nn=2)
poisson.set_grid_descriptor(gd)

poisson2 = FDPoissonSolver(nn=2, eps=1e-28)
poisson2.set_grid_descriptor(gd)

phi_g = gd.zeros()
phi2_g = gd.zeros()
rho_g = gd.zeros()
rho_g[4,5,6] = 1.0
rho_g[4,5,7] = -1.0

poisson.solve(phi_g, rho_g)
poisson2.solve(phi2_g, rho_g)
print("this", phi_g[4,5,:])
print("ref", phi2_g[4,5,:])
print("diff", phi_g[4,5,:]-phi2_g[4,5,:])

assert np.linalg.norm(phi_g-phi2_g) < 1e-10

gd = GridDescriptor((10,12,42), (10, 12, 42), pbc_c=(False,False,False))
poisson = NonPeriodicLauePoissonSolver(nn=1)
poisson.set_grid_descriptor(gd)
print("eigs", poisson.eigs_c[0])
poisson2 = FDPoissonSolver(nn=1, eps=1e-24)
poisson2.set_grid_descriptor(gd)

phi_g = gd.zeros()
phi2_g = gd.zeros()
rho_g = gd.zeros()
rho_g[:,:,1] = 1.0
rho_g[:,:,20] = -1.0
poisson.solve(phi_g, rho_g)
poisson2.solve(phi2_g, rho_g)
print("this", phi_g[4,5,:])
print("ref", phi2_g[4,5,:])
print("diff", phi_g[4,5,:]-phi2_g[4,5,:])

assert np.linalg.norm(phi_g-phi2_g) < 1e-10

X = idst2(dst2(rho_g))
diff = X-rho_g
print("dct diff", np.linalg.norm(diff))
assert np.linalg.norm(diff) < 1e-13



# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.

"""Atomic-centered localized functions."""

from math import pi

import numpy as np

from gpaw import debug
from gpaw.utilities import is_contiguous
from gpaw.mpi import run
import _gpaw


MASTER = 0


def create_localized_functions(functions, gd, spos_c,
                               dtype=float, cut=False,
                               forces=True, lfbc=None):
    """Create LocFuncs object.

    From a list of splines, a grid-descriptor and a scaled position,
    create a LocFuncs object.  If this domain does not contribute to
    the localized functions, None is returned.

    Parameters
    ==========
    dtype: float or complex
        Type of arrays to operate on.
    cut: bool
        Allow functions to cut boundaries when not periodic.
    forces: bool
        Calculate derivatives.
    lfbc: LocFuncBroadcaster
        Parallelization over **k**-points/spins.
    """

    lfs = LocFuncs(functions, gd, spos_c,
                   dtype, cut, forces, lfbc)

    if len(lfs.box_b) > 0:
        return lfs
    else:
        # No boxes in this domain:
        return None


class LocFuncs:
    """Class to handle atomic-centered localized functions."""
    def __init__(self, functions, gd, spos_c,
                 dtype, cut, forces, lfbc):
        """Create LocFuncs object.

        Use create_localized_functions() to create this object."""

        # Quick hack to take care of some special cases
        # (hgh setups without projectors and/or pseudo core density)
        if functions[0] is None:
            self.box_b = []
            return

        # We assume that all functions have the same cut-off:
        rcut = functions[0].get_cutoff()

        box_b = gd.get_boxes(spos_c, rcut, cut)

        self.box_b = []
        self.sdisp_bc = np.zeros((len(box_b), 3))
        b = 0
        for beg_c, end_c, sdisp_c in box_b:
            box = LocalizedFunctions(functions, beg_c, end_c,
                                     spos_c, sdisp_c, gd,
                                     dtype, forces, lfbc)
            self.box_b.append(box)
            self.sdisp_bc[b] = sdisp_c
            b += 1

        self.ni = 0
        for radial in functions:
            l = radial.get_angular_momentum_number()
            assert l <= 4, 'C-code only does l <= 4.'
            self.ni += 2 * l + 1

        self.dtype = dtype
        self.comm = gd.comm
        self.phase_kb = None

        if gd.comm.size > 1:
            # Find out which ranks have a piece of the
            # localized functions:
            i_have_a_piece = np.array(int(b > 0))
            i_have_a_piece_r = np.empty(gd.comm.size, int)
            gd.comm.all_gather(i_have_a_piece, i_have_a_piece_r)
            if i_have_a_piece:
                self.ranks = np.arange(self.comm.size)[i_have_a_piece_r == 1]
                self.root = gd.get_rank_from_position(spos_c)
        else:
            self.ranks = [0]
            self.root = 0

    def set_phase_factors(self, k_kc):
        """Set phase factors for Bloch boundary conditions."""
        self.phase_kb = np.exp(2j * pi * np.inner(k_kc, self.sdisp_bc))

    def add(self, a_xg, coef_xi, k=None, communicate=False):
        """Add localized functions to extended arrays.

        Add the product of coef_xi and the localized functions to
        a_xg.  With Bloch boundary-condtions, k is used to
        index the phase-factors.  If communicate is True,
        coef_xi will be broadcasted from the root-CPU."""

        run(self.iadd(a_xg, coef_xi, k, communicate))

    def iadd(self, a_xg, coef_xi, k=None, communicate=False):
        """Iterator for adding localized functions to extended arrays."""
        if communicate:
            if len(self.ranks) == 1:
                # Nothing to do:
                yield None
            else:
                # Send coefficients to other ranks:
                if self.comm.rank == self.root:
                    requests = []
                    for rank in self.ranks:
                        if rank == self.root:
                            continue
                        request = self.comm.send(coef_xi, rank, 1329, False)
                        requests.append(request)
                    yield None
                    for request in requests:
                        self.comm.wait(request)
                else:
                    # Get coefficients from root:
                    shape = a_xg.shape[:-3] + (self.ni,)
                    coef_xi = np.zeros(shape, self.dtype)
                    request = self.comm.receive(coef_xi, self.root,
                                                1329, False)

                    yield None
                    self.comm.wait(request)
        yield None

        if k is None or self.phase_kb is None:
            # No k-points:
            for box in self.box_b:
                box.add(coef_xi, a_xg)
        else:
            # K-points:
            for box, phase in zip(self.box_b, self.phase_kb[k]):
                box.add(coef_xi / phase, a_xg)

        yield None

    def integrate(self, a_xg, result_xi, k=None):
        """Calculate integrals of arrays times localized functions.

        Return the integral of extended arrays times localized
        functions in result_xi.  Correct phase-factors are used if the
        **k**-point index k is not None (Bloch boundary-condtions)."""

        run(self.iintegrate(a_xg, result_xi, k))

    def iintegrate(self, a_xg, result_xi, k=None):
        """Iterator for projecting extended arrays onto localized functions.

        There are three steps (the iterator will yield three times):

        1. Root-cpu starts non-blocking receive.
        2. Do local integral.  Non-root cpus send results to root
           (non-blocking).
        3. Wait for sends/receives to finish, then root returns sum.
        """

        shape = a_xg.shape[:-3] + (self.ni,)
        tmp_xi = np.zeros(shape, self.dtype)

        if result_xi is None:
            result_xi = np.zeros(shape, self.dtype)

        isum = self.isum(result_xi)
        next(isum)
        yield None

        if k is None or self.phase_kb is None:
            # No k-points:
            for box in self.box_b:
                box.integrate(a_xg, tmp_xi)
                result_xi += tmp_xi
        else:
            # K-points:
            for box, phase in zip(self.box_b, self.phase_kb[k]):
                box.integrate(a_xg, tmp_xi)
                result_xi += phase * tmp_xi


        next(isum)
        yield None
        next(isum)
        yield None

    def sum(self, a_x, broadcast=False):
        """Sum up array.

        The default behavior is to let the owner-node return the
        result in a_x.  With broadcast=True, all nodes will return the
        result in a_x."""

        run(self.isum(a_x, broadcast))

    def isum(self, a_x, broadcast=False):
        """Iterator for adding arrays.

        There are three steps:

        1. Root-node starts receiving.
        2. Non-root nodes start sending.
        3. Wait.  Then root does sum.

        If broadcast is True, there will be two more steps:

        1. Root-node sends and non-root nodes receive.
        2. Wait.
        """

        ndomains = len(self.ranks)
        if ndomains > 1:
            if self.comm.rank == self.root:
                requests = []
                a_dx = np.empty((ndomains - 1,) + a_x.shape, self.dtype)
                d = 0
                for rank in self.ranks:
                    if rank == self.root:
                        continue
                    request = self.comm.receive(a_dx[d:d + 1], rank,
                                                1330, False)
                    requests.append(request)
                    d += 1

                yield None
                yield None

                for request in requests:
                    self.comm.wait(request)

                a_x += a_dx.sum(0)

                if broadcast:
                    yield None
                    requests = []
                    for rank in self.ranks:
                        if rank == self.root:
                            continue
                        request = self.comm.send(a_x, rank, 1331, False)
                        requests.append(request)
                    yield None
                    for request in requests:
                        self.comm.wait(request)
            else:
                yield None
                request = self.comm.send(a_x, self.root, 1330, False)
                yield None
                self.comm.wait(request)
                if broadcast:
                    yield None
                    request = self.comm.receive(a_x, self.root, 1331, False)
                    yield None
                    self.comm.wait(request)

        else:
            yield None
            yield None
            if broadcast:
                yield None
                yield None

        yield None

    def derivative(self, a_xg, result_xic, k=None):
        """Calculate derivatives of localized integrals.

        Return the *x*- *y*- and *z*-derivatives of the integral of
        extended arrays times localized functions in result_xi.
        Correct phase-factors are used if the **k**-point index k
        is not None (Block boundary-condtions)."""

        run(self.iderivative(a_xg, result_xic, k))

    def iderivative(self, a_xg, result_xic, k=None):
        """Iterator for calculating derivatives of localized integrals."""
        shape = a_xg.shape[:-3] + (self.ni, 3)
        tmp_xic = np.zeros(shape, self.dtype)
        if result_xic is None:
            result_xic = np.zeros(shape, self.dtype)

        isum = self.isum(result_xic)
        next(isum)
        yield None

        if k is None or self.phase_kb is None:
            # No k-points:
            for box in self.box_b:
                box.derivative(a_xg, tmp_xic)
                result_xic += tmp_xic
        else:
            # K-points:
            for box, phase in zip(self.box_b, self.phase_kb[k]):
                box.derivative(a_xg, tmp_xic)
                result_xic += phase * tmp_xic

        next(isum)
        yield None
        next(isum)
        yield None

    def add_density(self, n_G, f_i):
        """Add atomic electron density to extended density array.

        Special method for adding the atomic electron density
        calculated from atomic orbitals and occupation numbers
        f_i."""
        for box in self.box_b:
            box.add_density(n_G, f_i)

    def add_density2(self, n_G, D_p):
        """Add atomic electron density to extended density array.

        Special method for adding the atomic electron density
        calculated from all cross products of atomic orbitals
        weighted using the density matrix D_p.

        The method returns the integral of the atomic electron density
        """
        I = 0.0
        for box in self.box_b:
            I += box.add_density2(n_G, D_p)
        return I

    def norm(self):
        """Calculate norm of localized functions."""

        I_ic = np.zeros((self.ni, 4))
        for box in self.box_b:
            box.norm(I_ic)
        self.sum(I_ic, broadcast=True)
        return I_ic

    def normalize(self, I0):
        """Normalize localized functions.

        The integral of the first function (spherically symmetric, l =
        0) is normalized to the value I0 and the following
        functions (l > 0) are adjusted so that they integrate to
        zero."""

        I_ic = self.norm()
        for box in self.box_b:
            box.normalize(I0, I_ic)


class LocalizedFunctionsWrapper:
    """Python wrapper class for C-extension: LocalizedFunctions.

    This class is used for construction of the C-object and for adding
    type-checking to the C-methods."""

    def __init__(self, functions, beg_c, end_c, spos_c, sdisp_c, gd,
                 dtype, forces, locfuncbcaster):
        """Construct a LocalizedFunctions C-object.

        Evaluate function values from a list of splines
        (functions) inside a box between grid points beg_c
        (included) to end_c (not included).  The functions are
        centered at the scaled position spos_c displaced by
        sdisp_c (in units of lattice vectors), and gd is the
        grid-descriptor.

        Derivatives are calculated when forces=True."""

        assert dtype in [float, complex]

        # Who evaluates the function values?
        if locfuncbcaster is None:
            # I do!
            compute = True
        else:
            # One of the CPU's in the k-point communicator does it,
            # and will later broadcast to the others:
            compute = next(locfuncbcaster)

        size_c = end_c - beg_c
        corner_c = beg_c - gd.beg_c
        pos_c = (beg_c - (spos_c - sdisp_c) * gd.N_c) * gd.h_cv.diagonal()

        self.lfs = _gpaw.LocalizedFunctions(
            [function.spline for function in functions],
            size_c, gd.n_c, corner_c, gd.h_cv.diagonal().copy(), pos_c,
            dtype == float, forces, compute)

        if locfuncbcaster is not None:
            locfuncbcaster.add(self.lfs)

        self.ni = 0   # number of functions
        for function in functions:
            l = function.get_angular_momentum_number()
            self.ni += 2 * l + 1

        self.shape = tuple(gd.n_c)
        self.dtype = dtype
        self.forces = forces

    def integrate(self, a_xg, result_xi):
        """Calculate integrals of arrays times localized functions.

        Return the integral of extended arrays times localized
        functions in result_xi."""

        assert is_contiguous(a_xg, self.dtype)
        assert is_contiguous(result_xi, self.dtype)
        assert a_xg.shape[:-3] == result_xi.shape[:-1]
        assert a_xg.shape[-3:] == self.shape
        assert result_xi.shape[-1] == self.ni
        self.lfs.integrate(a_xg, result_xi)

    def derivative(self, a_xg, result_xic):
        """Calculate x-, y-, z-derivatives of localized integrals.

        Return the *x*- *y*- and *z*-derivatives of the integral of
        extended arrays times localized functions in
        result_xic."""

        assert self.forces
        assert is_contiguous(a_xg, self.dtype)
        assert is_contiguous(result_xic, self.dtype)
        assert a_xg.shape[:-3] == result_xic.shape[:-2]
        assert a_xg.shape[-3:] == self.shape
        assert result_xic.shape[-2:] == (self.ni, 3)
        self.lfs.derivative(a_xg, result_xic)

    def add(self, coef_xi, a_xg):
        """Add localized functions to extended arrays.

        Add the product of coef_xi and the localized functions to
        a_xg."""

        assert is_contiguous(a_xg, self.dtype)
        assert is_contiguous(coef_xi, self.dtype)
        assert a_xg.shape[:-3] == coef_xi.shape[:-1]
        assert a_xg.shape[-3:] == self.shape
        assert coef_xi.shape[-1] == self.ni
        self.lfs.add(coef_xi, a_xg)

    def add_density(self, n_G, f_i):
        """Add atomic electron density to extended density array.

        Special method for adding the atomic electron density
        calculated from atomic orbitals and occupation numbers
        f_i."""

        assert is_contiguous(n_G, float)
        assert is_contiguous(f_i, float)
        assert n_G.shape == self.shape
        assert f_i.shape == (self.ni,)
        self.lfs.add_density(n_G, f_i)

    def add_density2(self, n_G, D_p):
        """Add atomic electron density to extended density array.

        Special method for adding the atomic electron density
        calculated from all cross products of atomic orbitals
        weighted using the density matrix D_p.

        The method returns the integral of the atomic electron density
        """

        assert is_contiguous(n_G, float)
        assert is_contiguous(D_p, float)
        assert n_G.shape == self.shape
        assert D_p.shape == (self.ni * (self.ni + 1) / 2,)
        return self.lfs.add_density2(n_G, D_p)

    def norm(self, I_ic):
        """Integrate functions."""
        assert I_ic.flags.contiguous
        assert I_ic.dtype == float
        assert I_ic.shape == (self.ni, 4)
        return self.lfs.norm(I_ic)

    def normalize(self, I0, I_ic):
        """Normalize functions."""
        assert I_ic.flags.contiguous
        assert I_ic.dtype == float
        assert I_ic.shape == (self.ni, 4)
        return self.lfs.normalize(I0, I_ic)

    def get_functions(self):
        """Return functions in ndarray."""
        return self.lfs.get_functions()

    def set_corner(self, start_c):
        """Return functions in ndarray."""
        assert start_c.flags.contiguous
        assert start_c.dtype == int
        assert start_c.shape == (3,)
        self.lfs.set_corner(start_c)


if debug:
    # Add type and sanity checks:
    LocalizedFunctions = LocalizedFunctionsWrapper
else:
    # Just use the bare C-object for efficiency:
    def LocalizedFunctions(functions, beg_c, end_c, spos_c, sdisp_c, gd,
                           dtype, forces, locfuncbcaster):
        return LocalizedFunctionsWrapper(functions, beg_c, end_c, spos_c,
                                         sdisp_c, gd,
                                         dtype, forces, locfuncbcaster).lfs


class LocFuncBroadcaster:
    """Parallelize evaluation of localized functions over k-points/spins."""
    def __init__(self, kpt_comm):
        """Set up framework for parallelization over k-points/spins.

        One of the CPU's in the k-point communicator will do the work
        and later broadcast arrays to the other CPU's.
        """

        self.kpt_comm = kpt_comm
        self.size = kpt_comm.size
        self.rank = kpt_comm.rank
        self.reset()

    def reset(self):
        self.lfs = []
        self.root = 0

    def __next__(self):
        """Distribute work.

        Return True if this CPU should do anything.
        """

        compute = (self.root == self.rank)
        self.root = (self.root + 1) % self.size
        return compute

    next = __next__  # for Python 2

    def add(self, lf):
        """Add localized functions to pool."""
        self.lfs.append(lf)

    def broadcast(self):
        """Distribute arrays to all processors."""
        if self.size > 1:
            for root, lf in enumerate(self.lfs):
                lf.broadcast(self.kpt_comm, root % self.size)
        self.reset()



from __future__ import print_function, division

import functools
import numbers
import sys
from math import pi

import numpy as np
from scipy.spatial import Delaunay, cKDTree

from ase.units import Hartree
from ase.utils import convert_string_to_fd
from ase.utils.timing import timer, Timer

import gpaw.mpi as mpi
from gpaw import GPAW
from gpaw.fd_operators import Gradient
from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.occupations import FermiDirac
from gpaw.response.math_func import (two_phi_planewave_integrals,
                                     two_phi_nabla_planewave_integrals)
from gpaw.utilities.blas import gemm
from gpaw.utilities.progressbar import ProgressBar
from gpaw.wavefunctions.pw import PWLFC
from gpaw.bztools import get_reduced_bz, unique_rows


class KPoint:
    def __init__(self, s, K, n1, n2, blocksize, na, nb,
                 ut_nR, eps_n, f_n, P_ani, shift_c):
        self.s = s    # spin index
        self.K = K    # BZ k-point index
        self.n1 = n1  # first band
        self.n2 = n2  # first band not included
        self.blocksize = blocksize
        self.na = na  # first band of block
        self.nb = nb  # first band of block not included
        self.ut_nR = ut_nR      # periodic part of wave functions in real-space
        self.eps_n = eps_n      # eigenvalues
        self.f_n = f_n          # occupation numbers
        self.P_ani = P_ani      # PAW projections
        self.shift_c = shift_c  # long story - see the
        # PairDensity.construct_symmetry_operators() method


class KPointPair:
    """This class defines the kpoint-pair container object.

    Used for calculating pair quantities it contains two kpoints,
    and an associated set of Fourier components."""
    def __init__(self, kpt1, kpt2, Q_G):
        self.kpt1 = kpt1
        self.kpt2 = kpt2
        self.Q_G = Q_G

    def get_k1(self):
        """ Return KPoint object 1."""
        return self.kpt1

    def get_k2(self):
        """ Return KPoint object 2."""
        return self.kpt2

    def get_planewave_indices(self):
        """ Return the planewave indices associated with this pair."""
        return self.Q_G

    def get_transition_energies(self, n_n, m_m):
        """Return the energy difference for specified bands."""
        n_n = np.array(n_n)
        m_m = np.array(m_m)
        kpt1 = self.kpt1
        kpt2 = self.kpt2
        deps_nm = (kpt1.eps_n[n_n - self.kpt1.n1][:, np.newaxis] -
                   kpt2.eps_n[m_m - self.kpt2.n1])
        return deps_nm

    def get_occupation_differences(self, n_n, m_m):
        """Get difference in occupation factor between specified bands."""
        n_n = np.array(n_n)
        m_m = np.array(m_m)
        kpt1 = self.kpt1
        kpt2 = self.kpt2
        df_nm = (kpt1.f_n[n_n - self.kpt1.n1][:, np.newaxis] -
                 kpt2.f_n[m_m - self.kpt2.n1])
        return df_nm


class PWSymmetryAnalyzer:
    """Class for handling planewave symmetries."""
    def __init__(self, kd, pd, txt=sys.stdout,
                 disable_point_group=False,
                 disable_non_symmorphic=True,
                 disable_time_reversal=False,
                 timer=None):
        """Creates a PWSymmetryAnalyzer object.

        Determines which of the symmetries of the atomic structure
        that is compatible with the reciprocal lattice. Contains the
        necessary functions for mapping quantities between kpoints,
        and or symmetrizing arrays.

        kd: KPointDescriptor
            The kpoint descriptor containing the
            information about symmetries and kpoints.
        pd: PWDescriptor
            Plane wave descriptor that contains the reciprocal
            lattice .
        txt: str
            Output file.
        disable_point_group: bool
            Switch for disabling point group symmetries.
        disable_non_symmorphic:
            Switch for disabling non symmorphic symmetries.
        disable_time_reversal:
            Switch for disabling time reversal.
        """
        self.pd = pd
        self.kd = kd
        self.fd = txt

        # Caveats
        assert disable_non_symmorphic, \
            print('You are not allowed to use non symmorphic syms, sorry. ',
                  file=self.fd)

        # Settings
        self.disable_point_group = disable_point_group
        self.disable_time_reversal = disable_time_reversal
        self.disable_non_symmorphic = disable_non_symmorphic
        if (kd.symmetry.has_inversion or not kd.symmetry.time_reversal) and \
           not self.disable_time_reversal:
            print('\nThe ground calculation does not support time-reversal ' +
                  'symmetry possibly because it has an inversion center ' +
                  'or that it has been manually deactivated. \n', file=self.fd)
            self.disable_time_reversal = True

        self.disable_symmetries = (self.disable_point_group and
                                   self.disable_time_reversal and
                                   self.disable_non_symmorphic)

        # Number of symmetries
        U_scc = kd.symmetry.op_scc
        self.nU = len(U_scc)

        self.nsym = 2 * self.nU
        self.use_time_reversal = not self.disable_time_reversal

        # Which timer to use
        self.timer = timer or Timer()

        self.KDTree = cKDTree(np.mod(np.mod(kd.bzk_kc, 1).round(6), 1))

        # Initialize
        self.initialize()

    @timer('Initialize')
    def initialize(self):
        """Initialize relevant quantities."""
        self.infostring = ''
        if self.disable_point_group:
            self.infostring += 'Point group not included. '
        else:
            self.infostring += 'Point group included. '

        if self.disable_time_reversal:
            self.infostring += 'Time reversal not included. '
        else:
            self.infostring += 'Time reversal included. '

        if self.disable_non_symmorphic:
            self.infostring += 'Disabled non symmorphic symmetries. '
        else:
            self.infostring += 'Time reversal included. '

        if self.disable_symmetries:
            self.infostring += 'All symmetries have been disabled. '

        # Do the work
        self.analyze_symmetries()
        self.analyze_kpoints()
        self.initialize_G_maps()

        # Print info
        print(self.infostring, file=self.fd)
        self.print_symmetries()

    def find_kpoint(self, k_c):
        return self.KDTree.query(np.mod(np.mod(k_c, 1).round(6), 1))[1]

    def print_symmetries(self):
        """Handsome print function for symmetry operations."""

        p = functools.partial(print, file=self.fd)

        p()
        nx = 6 if self.disable_non_symmorphic else 3
        ns = len(self.s_s)
        y = 0
        for y in range((ns + nx - 1) // nx):
            for c in range(3):
                for x in range(nx):
                    s = x + y * nx
                    if s == ns:
                        break
                    tmp = self.get_symmetry_operator(self.s_s[s])
                    op_cc, sign, TR, shift_c, ft_c = tmp
                    op_c = sign * op_cc[c]
                    p('  (%2d %2d %2d)' % tuple(op_c), end='')
                p()
            p()

    @timer('Analyze')
    def analyze_kpoints(self):
        """Calculate the reduction in the number of kpoints."""
        K_gK = self.group_kpoints()
        ng = len(K_gK)
        self.infostring += '{0} groups of equivalent kpoints. '.format(ng)
        percent = (1. - (ng + 0.) / self.kd.nbzkpts) * 100
        self.infostring += '{0}% reduction. '.format(percent)

    @timer('Analyze symmetries.')
    def analyze_symmetries(self):
        """Determine allowed symmetries.

        An direct symmetry U must fulfill::

          U \mathbf{q} = q + \Delta

        Under time-reversal (indirect) it must fulfill::

          -U \mathbf{q} = q + \Delta

        where :math:`\Delta` is a reciprocal lattice vector.
        """
        pd = self.pd

        # Shortcuts
        q_c = pd.kd.bzk_kc[0]
        kd = self.kd

        U_scc = kd.symmetry.op_scc
        nU = self.nU
        nsym = self.nsym

        shift_sc = np.zeros((nsym, 3), int)
        conserveq_s = np.zeros(nsym, bool)

        newq_sc = np.dot(U_scc, q_c)

        # Direct symmetries
        dshift_sc = (newq_sc - q_c[np.newaxis]).round().astype(int)
        inds_s = np.argwhere((newq_sc == q_c[np.newaxis] + dshift_sc).all(1))
        conserveq_s[inds_s] = True

        shift_sc[:nU] = dshift_sc

        # Time reversal
        trshift_sc = (-newq_sc - q_c[np.newaxis]).round().astype(int)
        trinds_s = np.argwhere((-newq_sc == q_c[np.newaxis] +
                                trshift_sc).all(1)) + nU
        conserveq_s[trinds_s] = True
        shift_sc[nU:nsym] = trshift_sc

        # The indices of the allowed symmetries
        s_s = conserveq_s.nonzero()[0]

        # Filter out disabled symmetries
        if self.disable_point_group:
            s_s = list(filter(self.is_not_point_group, s_s))

        if self.disable_time_reversal:
            s_s = list(filter(self.is_not_time_reversal, s_s))

        if self.disable_non_symmorphic:
            s_s = list(filter(self.is_not_non_symmorphic, s_s))

#        stmp_s = []
#        for s in s_s:
#            if self.kd.bz2bz_ks[0, s] == -1:
#                assert (self.kd.bz2bz_ks[:, s] == -1).all()
#            else:
#                stmp_s.append(s)

#        s_s = stmp_s

        self.infostring += 'Found {} allowed symmetries. '.format(len(s_s))
        self.s_s = s_s
        self.shift_sc = shift_sc

    def is_not_point_group(self, s):
        U_scc = self.kd.symmetry.op_scc
        nU = self.nU
        return (U_scc[s % nU] == np.eye(3)).all()

    def is_not_time_reversal(self, s):
        nU = self.nU
        return not bool(s // nU)

    def is_not_non_symmorphic(self, s):
        ft_sc = self.kd.symmetry.ft_sc
        nU = self.nU
        return not bool(ft_sc[s % nU].any())

    def how_many_symmetries(self):
        """Return number of symmetries."""
        return len(self.s_s)

    @timer('Group kpoints')
    def group_kpoints(self, K_k=None):
        """Group kpoints according to the reduced symmetries"""
        if K_k is None:
            K_k = np.arange(self.kd.nbzkpts)
        s_s = self.s_s
        bz2bz_ks = self.kd.bz2bz_ks
        nk = len(bz2bz_ks)
        sbz2sbz_ks = bz2bz_ks[K_k][:, s_s]  # Reduced number of symmetries
        # Avoid -1 (see documentation in gpaw.symmetry)
        sbz2sbz_ks[sbz2sbz_ks == -1] = nk

        smallestk_k = np.sort(sbz2sbz_ks)[:, 0]
        k2g_g = np.unique(smallestk_k, return_index=True)[1]

        K_gs = sbz2sbz_ks[k2g_g]
        K_gk = [np.unique(K_s[K_s != nk]) for K_s in K_gs]

        return K_gk

    def get_BZ(self):
        # Get the little group of q
        U_scc = []
        for s in self.s_s:
            U_cc, sign, _, _, _ = self.get_symmetry_operator(s)
            U_scc.append(sign * U_cc)
        U_scc = np.array(U_scc)

        # Determine the irreducible BZ
        bzk_kc, ibzk_kc = get_reduced_bz(self.pd.gd.cell_cv,
                                         U_scc,
                                         False)

        return bzk_kc

    def get_reduced_kd(self, pbc_c=np.ones(3, bool)):
        # Get the little group of q
        U_scc = []
        for s in self.s_s:
            U_cc, sign, _, _, _ = self.get_symmetry_operator(s)
            U_scc.append(sign * U_cc)
        U_scc = np.array(U_scc)

        # Determine the irreducible BZ
        bzk_kc, ibzk_kc = get_reduced_bz(self.pd.gd.cell_cv,
                                         U_scc,
                                         False,
                                         pbc_c=pbc_c)

        n = 3
        N_xc = np.indices((n, n, n)).reshape((3, n**3)).T - n // 2

        # Find the irreducible kpoints
        tess = Delaunay(ibzk_kc)
        ik_kc = []
        for N_c in N_xc:
            k_kc = self.kd.bzk_kc + N_c
            k_kc = k_kc[tess.find_simplex(k_kc) >= 0]
            if not len(ik_kc) and len(k_kc):
                ik_kc = unique_rows(k_kc)
            elif len(k_kc):
                ik_kc = unique_rows(np.append(k_kc, ik_kc, axis=0))

        return KPointDescriptor(ik_kc)

    def unfold_kpoints(self, points_pv, tol=1e-8, mod=None):
        points_pc = np.dot(points_pv, self.pd.gd.cell_cv.T) / (2 * np.pi)

        # Get the little group of q
        U_scc = []
        for s in self.s_s:
            U_cc, sign, _, _, _ = self.get_symmetry_operator(s)
            U_scc.append(sign * U_cc)
        U_scc = np.array(U_scc)

        points = np.concatenate(np.dot(points_pc, U_scc.transpose(0, 2, 1)))
        points = unique_rows(points, tol=tol, mod=mod)
        points = np.dot(points, self.pd.gd.icell_cv) * (2 * np.pi)
        return points

    def get_kpoint_weight(self, k_c):
        K = self.find_kpoint(k_c)
        iK = self.kd.bz2ibz_k[K]
        K_k = self.unfold_ibz_kpoint(iK)
        K_gK = self.group_kpoints(K_k)

        for K_k in K_gK:
            if K in K_k:
                if self.kd.refine_info is not None:
                    weight = sum(self.kd.refine_info.weight_k[K_k])
                    return weight
                else:
                    return len(K_k)

    def get_kpoint_mapping(self, K1, K2):
        """Get index of symmetry for mapping between K1 and K2"""
        s_s = self.s_s
        bz2bz_ks = self.kd.bz2bz_ks
        bzk2rbz_s = bz2bz_ks[K1][s_s]
        try:
            s = np.argwhere(bzk2rbz_s == K2)[0][0]
        except IndexError:
            print('K = {0} cannot be mapped into K = {1}'.format(K1, K2),
                  file=self.fd)
            raise
        return s_s[s]

    def get_shift(self, K1, K2, U_cc, sign):
        """Get shift for mapping between K1 and K2."""
        kd = self.kd
        k1_c = kd.bzk_kc[K1]
        k2_c = kd.bzk_kc[K2]

        shift_c = np.dot(U_cc, k1_c) - k2_c * sign
        assert np.allclose(shift_c.round(), shift_c)
        shift_c = shift_c.round().astype(int)

        return shift_c

    @timer('map_G')
    def map_G(self, K1, K2, a_MG):
        """Map a function of G from K1 to K2. """
        if len(a_MG) == 0:
            return []

        if K1 == K2:
            return a_MG

        G_G, sign = self.map_G_vectors(K1, K2)

        s = self.get_kpoint_mapping(K1, K2)
        U_cc, _, TR, shift_c, ft_c = self.get_symmetry_operator(s)

        return TR(a_MG[..., G_G])

    def symmetrize_wGG(self, A_wGG):
        """Symmetrize an array in GG'."""

        for A_GG in A_wGG:
            tmp_GG = np.zeros_like(A_GG)

            for s in self.s_s:
                G_G, sign, _ = self.G_sG[s]
                if sign == 1:
                    tmp_GG += A_GG[G_G, :][:, G_G]
                if sign == -1:
                    tmp_GG += A_GG[G_G, :][:, G_G].T

            A_GG[:] = tmp_GG / self.how_many_symmetries()

    def symmetrize_wxx(self, A_wxx, optical_limit=False):
        """Symmetrize an array in xx'."""
        tmp_wxx = np.zeros_like(A_wxx)

        A_cv = self.pd.gd.cell_cv
        iA_cv = self.pd.gd.icell_cv

        if self.use_time_reversal:
            AT_wxx = np.transpose(A_wxx, (0, 2, 1))

        for s in self.s_s:
            G_G, sign, shift_c = self.G_sG[s]
            if optical_limit:
                G_G = np.array(G_G) + 2
                G_G = np.insert(G_G, 0, [0, 1])
                U_cc, _, TR, shift_c, ft_c = self.get_symmetry_operator(s)
                M_vv = np.dot(np.dot(A_cv.T, U_cc.T), iA_cv)

            if sign == 1:
                tmp = A_wxx[:, G_G, :][:, :, G_G]
                if optical_limit:
                    tmp[:, 0:3, :] = np.transpose(np.dot(M_vv.T,
                                                         tmp[:, 0:3, :]),
                                                  (1, 0, 2))
                    tmp[:, :, 0:3] = np.dot(tmp[..., 0:3], M_vv)
                tmp_wxx += tmp
            elif sign == -1:
                tmp = AT_wxx[:, G_G, :][:, :, G_G]
                if optical_limit:
                    tmp[:, 0:3, :] = np.transpose(np.dot(M_vv.T,
                                                         tmp[:, 0:3, :]),
                                                  (1, 0, 2)) * sign
                    tmp[:, :, 0:3] = np.dot(tmp[:, :, 0:3], M_vv) * sign
                tmp_wxx += tmp

        # Inplace overwriting
        A_wxx[:] = tmp_wxx / self.how_many_symmetries()

    def symmetrize_wxvG(self, A_wxvG):
        """Symmetrize chi0_wxvG"""
        A_cv = self.pd.gd.cell_cv
        iA_cv = self.pd.gd.icell_cv

        if self.use_time_reversal:
            # ::-1 corresponds to transpose in wing indices
            AT_wxvG = A_wxvG[:, ::-1]

        tmp_wxvG = np.zeros_like(A_wxvG)
        for s in self.s_s:
            G_G, sign, shift_c = self.G_sG[s]
            U_cc, _, TR, shift_c, ft_c = self.get_symmetry_operator(s)
            M_vv = np.dot(np.dot(A_cv.T, U_cc.T), iA_cv)
            if sign == 1:
                tmp = sign * np.dot(M_vv.T, A_wxvG[..., G_G])
            elif sign == -1:
                tmp = sign * np.dot(M_vv.T, AT_wxvG[..., G_G])
            tmp_wxvG += np.transpose(tmp, (1, 2, 0, 3))

        # Overwrite the input
        A_wxvG[:] = tmp_wxvG / self.how_many_symmetries()

    def symmetrize_wvv(self, A_wvv):
        """Symmetrize chi_wvv."""
        A_cv = self.pd.gd.cell_cv
        iA_cv = self.pd.gd.icell_cv
        tmp_wvv = np.zeros_like(A_wvv)
        if self.use_time_reversal:
            AT_wvv = np.transpose(A_wvv, (0, 2, 1))

        for s in self.s_s:
            G_G, sign, shift_c = self.G_sG[s]
            U_cc, _, TR, shift_c, ft_c = self.get_symmetry_operator(s)
            M_vv = np.dot(np.dot(A_cv.T, U_cc.T), iA_cv)
            if sign == 1:
                tmp = np.dot(np.dot(M_vv.T, A_wvv), M_vv)
            elif sign == -1:
                tmp = np.dot(np.dot(M_vv.T, AT_wvv), M_vv)
            tmp_wvv += np.transpose(tmp, (1, 0, 2))

        # Overwrite the input
        A_wvv[:] = tmp_wvv / self.how_many_symmetries()

    @timer('map_v')
    def map_v(self, K1, K2, a_Mv):
        """Map a function of v (cartesian component) from K1 to K2."""

        if len(a_Mv) == 0:
            return []

        if K1 == K2:
            return a_Mv

        A_cv = self.pd.gd.cell_cv
        iA_cv = self.pd.gd.icell_cv

        # Get symmetry
        s = self.get_kpoint_mapping(K1, K2)
        U_cc, sign, TR, _, ft_c = self.get_symmetry_operator(s)

        # Create cartesian operator
        M_vv = np.dot(np.dot(A_cv.T, U_cc.T), iA_cv)
        return sign * np.dot(TR(a_Mv), M_vv)

    def timereversal(self, s):
        """Is this a time-reversal symmetry?"""
        tr = bool(s // self.nU)
        return tr

    def get_symmetry_operator(self, s):
        """Return symmetry operator s."""
        U_scc = self.kd.symmetry.op_scc
        ft_sc = self.kd.symmetry.op_scc

        reds = s % self.nU
        if self.timereversal(s):
            TR = lambda x: x.conj()
            sign = -1
        else:
            sign = 1
            TR = lambda x: x

        return U_scc[reds], sign, TR, self.shift_sc[s], ft_sc[reds]

    @timer('map_G_vectors')
    def map_G_vectors(self, K1, K2):
        """Return G vector mapping."""
        s = self.get_kpoint_mapping(K1, K2)
        G_G, sign, shift_c = self.G_sG[s]

        return G_G, sign

    def initialize_G_maps(self):
        """Calculate the Gvector mappings."""
        pd = self.pd
        B_cv = 2.0 * np.pi * pd.gd.icell_cv
        G_Gv = pd.get_reciprocal_vectors(add_q=False)
        G_Gc = np.dot(G_Gv, np.linalg.inv(B_cv))
        Q_G = pd.Q_qG[0]

        G_sG = [None] * self.nsym
        UG_sGc = [None] * self.nsym
        Q_sG = [None] * self.nsym
        for s in self.s_s:
            U_cc, sign, TR, shift_c, ft_c = self.get_symmetry_operator(s)
            iU_cc = np.linalg.inv(U_cc).T
            UG_Gc = np.dot(G_Gc - shift_c, sign * iU_cc)

            assert np.allclose(UG_Gc.round(), UG_Gc)
            UQ_G = np.ravel_multi_index(UG_Gc.round().astype(int).T,
                                        pd.gd.N_c, 'wrap')

            G_G = len(Q_G) * [None]
            for G, UQ in enumerate(UQ_G):
                try:
                    G_G[G] = np.argwhere(Q_G == UQ)[0][0]
                except IndexError:
                    print('This should not be possible but' +
                          'a G-vector was mapped outside the sphere')
                    raise IndexError
            UG_sGc[s] = UG_Gc
            Q_sG[s] = UQ_G
            G_sG[s] = [G_G, sign, shift_c]
        self.G_Gc = G_Gc
        self.UG_sGc = UG_sGc
        self.Q_sG = Q_sG
        self.G_sG = G_sG

    def unfold_ibz_kpoint(self, ik):
        """Return kpoints related to irreducible kpoint."""
        kd = self.kd
        K_k = np.unique(kd.bz2bz_ks[kd.ibz2bz_k[ik]])
        K_k = K_k[K_k != -1]
        return K_k


class PairDensity:
    def __init__(self, calc, ecut=50, response='density',
                 ftol=1e-6, threshold=1,
                 real_space_derivatives=False,
                 world=mpi.world, txt='-', timer=None,
                 nblocks=1, gate_voltage=None):

        if ecut is not None:
            ecut /= Hartree

        if gate_voltage is not None:
            gate_voltage = gate_voltage / Hartree
        
        self.response = response
        self.ecut = ecut
        self.ftol = ftol
        self.threshold = threshold
        self.real_space_derivatives = real_space_derivatives
        self.gate_voltage = gate_voltage

        self.timer = timer or Timer()

        self.world = world

        if nblocks == 1:
            self.blockcomm = world.new_communicator([world.rank])
            self.kncomm = world
        else:
            assert world.size % nblocks == 0, world.size
            rank1 = world.rank // nblocks * nblocks
            rank2 = rank1 + nblocks
            self.blockcomm = self.world.new_communicator(range(rank1, rank2))
            ranks = range(world.rank % nblocks, world.size, nblocks)
            self.kncomm = self.world.new_communicator(ranks)

        self.fd = convert_string_to_fd(txt, world)

        with self.timer('Read ground state'):
            if isinstance(calc, str):
                print('Reading ground state calculation:\n  %s' % calc,
                      file=self.fd)
                calc = GPAW(calc, txt=None, communicator=mpi.serial_comm)
            else:
                assert calc.wfs.world.size == 1

        assert calc.wfs.kd.symmetry.symmorphic
        self.calc = calc

        self.fermi_level = self.calc.occupations.get_fermi_level()
        
        if gate_voltage is not None:
            self.add_gate_voltage(gate_voltage)
        else:
            self.add_gate_voltage(gate_voltage=0)

        self.spos_ac = calc.spos_ac

        self.nocc1 = None  # number of completely filled bands
        self.nocc2 = None  # number of non-empty bands
        self.count_occupied_bands()

        self.ut_sKnvR = None  # gradient of wave functions for optical limit

        self.vol = abs(np.linalg.det(calc.wfs.gd.cell_cv))

        kd = self.calc.wfs.kd
        self.KDTree = cKDTree(np.mod(np.mod(kd.bzk_kc, 1).round(6), 1))
        print('Number of blocks:', nblocks, file=self.fd)

    def find_kpoint(self, k_c):
        return self.KDTree.query(np.mod(np.mod(k_c, 1).round(6), 1))[1]

    def add_gate_voltage(self, gate_voltage=0):
        """Shifts the Fermi-level by e * Vg. By definition e = 1."""
        assert isinstance(self.calc.occupations, FermiDirac)
        print('Shifting Fermi-level by %.2f eV' % (gate_voltage * Hartree),
              file=self.fd)
        self.fermi_level += gate_voltage
        for kpt in self.calc.wfs.kpt_u:
            kpt.f_n = (self.shift_occupations(kpt.eps_n, gate_voltage) *
                       kpt.weight)

    def shift_occupations(self, eps_n, gate_voltage):
        """Shift fermilevel."""
        fermi = self.fermi_level
        width = self.calc.occupations.width
        if width < 1e-9:
            return (eps_n < fermi).astype(float)
        else:
            tmp = (eps_n - fermi) / width
        f_n = np.zeros_like(eps_n)
        f_n[tmp <= 100] = 1 / (1 + np.exp(tmp[tmp <= 100]))
        f_n[tmp > 100] = 0.0
        return f_n

    def count_occupied_bands(self):
        self.nocc1 = 9999999
        self.nocc2 = 0
        for kpt in self.calc.wfs.kpt_u:
            f_n = kpt.f_n / kpt.weight
            self.nocc1 = min((f_n > 1 - self.ftol).sum(), self.nocc1)
            self.nocc2 = max((f_n > self.ftol).sum(), self.nocc2)
        print('Number of completely filled bands:', self.nocc1, file=self.fd)
        print('Number of partially filled bands:', self.nocc2, file=self.fd)
        print('Total number of bands:', self.calc.wfs.bd.nbands,
              file=self.fd)

    def distribute_k_points_and_bands(self, band1, band2, kpts=None):
        """Distribute spins, k-points and bands.

        nbands: int
            Number of bands for each spin/k-point combination.

        The attribute self.mysKn1n2 will be set to a list of (s, K, n1, n2)
        tuples that this process handles.
        """

        wfs = self.calc.wfs

        if kpts is None:
            kpts = np.arange(wfs.kd.nbzkpts)

        nbands = band2 - band1
        size = self.kncomm.size
        rank = self.kncomm.rank
        ns = wfs.nspins
        nk = len(kpts)
        n = (ns * nk * nbands + size - 1) // size
        i1 = rank * n
        i2 = min(i1 + n, ns * nk * nbands)

        self.mysKn1n2 = []
        i = 0
        for s in range(ns):
            for K in kpts:
                n1 = min(max(0, i1 - i), nbands)
                n2 = min(max(0, i2 - i), nbands)
                if n1 != n2:
                    self.mysKn1n2.append((s, K, n1 + band1, n2 + band1))
                i += nbands

        print('BZ k-points:', self.calc.wfs.kd, file=self.fd)
        print('Distributing spins, k-points and bands (%d x %d x %d)' %
              (ns, nk, nbands),
              'over %d process%s' %
              (self.kncomm.size, ['es', ''][self.kncomm.size == 1]),
              file=self.fd)
        print('Number of blocks:', self.blockcomm.size, file=self.fd)

    @timer('Get a k-point')
    def get_k_point(self, s, k_c, n1, n2, load_wfs=True, block=False):
        """Return wave functions for a specific k-point and spin.

        s: int
            Spin index (0 or 1).
        K: int
            BZ k-point index.
        n1, n2: int
            Range of bands to include.
        """

        wfs = self.calc.wfs
        kd = wfs.kd

        # Parse kpoint: is k_c an index or a vector
        if not isinstance(k_c, numbers.Integral):
            K = self.find_kpoint(k_c)
            shift0_c = (kd.bzk_kc[K] - k_c).round().astype(int)
        else:
            # Fall back to index
            K = k_c
            shift0_c = np.array([0, 0, 0])
            k_c = None

        if block:
            nblocks = self.blockcomm.size
            rank = self.blockcomm.rank
        else:
            nblocks = 1
            rank = 0

        blocksize = (n2 - n1 + nblocks - 1) // nblocks
        na = min(n1 + rank * blocksize, n2)
        nb = min(na + blocksize, n2)

        U_cc, T, a_a, U_aii, shift_c, time_reversal = \
            self.construct_symmetry_operators(K, k_c=k_c)

        shift_c += -shift0_c
        ik = wfs.kd.bz2ibz_k[K]
        kpt = wfs.kpt_u[s * wfs.kd.nibzkpts + ik]

        assert n2 <= len(kpt.eps_n), \
            'Increase GS-nbands or decrease chi0-nbands!'
        eps_n = kpt.eps_n[n1:n2]
        f_n = kpt.f_n[n1:n2] / kpt.weight
        
        if not load_wfs:
            return KPoint(s, K, n1, n2, blocksize, na, nb,
                          None, eps_n, f_n, None, shift_c)

        with self.timer('load wfs'):
            psit_nG = kpt.psit_nG
            ut_nR = wfs.gd.empty(nb - na, wfs.dtype)
            for n in range(na, nb):
                ut_nR[n - na] = T(wfs.pd.ifft(psit_nG[n], ik))

        with self.timer('Load projections'):
            P_ani = []
            for b, U_ii in zip(a_a, U_aii):
                P_ni = np.dot(kpt.P_ani[b][na:nb], U_ii)
                if time_reversal:
                    P_ni = P_ni.conj()
                P_ani.append(P_ni)

        return KPoint(s, K, n1, n2, blocksize, na, nb,
                      ut_nR, eps_n, f_n, P_ani, shift_c)

    def generate_pair_densities(self, pd, m1, m2, spins, intraband=True,
                                PWSA=None, disable_optical_limit=False,
                                unsymmetrized=False, use_more_memory=1):
        """Generator for returning pair densities.

        Returns the pair densities between the occupied and
        the states in range(m1, m2).

        pd: PWDescriptor
            Plane-wave descriptor for a single q-point.
        m1: int
            Index of first unoccupied band.
        m2: int
            Index of last unoccupied band.
        spins: list
            List of spin indices included.
        intraband: bool
            Include intraband transitions in optical limit.
        PWSA: PlanewaveSymmetryAnalyzer
            If supplied uses this object to determine the symmetries
            of the pair-densities.
        disable_optical_limit: bool
            Disable optical limit.
        unsymmetrized: bool
            Only return pair-densities from one kpoint in each
            group of equivalent kpoints.
        use_more_memory: float
            Group more pair densities for several occupied bands
            together before returning. Here 0 <= use_more_memory <= 1,
            where zero is the minimal amount of memory, and 1 is the maximal.
        """
        assert 0 <= use_more_memory <= 1

        q_c = pd.kd.bzk_kc[0]
        optical_limit = np.allclose(q_c, 0.0) and self.response == 'density'
        optical_limit = not disable_optical_limit and optical_limit

        Q_aGii = self.initialize_paw_corrections(pd)
        self.Q_aGii = Q_aGii  # This is used in g0w0

        if PWSA is None:
            with self.timer('Symmetry analyzer'):
                PWSA = PWSymmetryAnalyzer  # Line too long otherwise
                PWSA = PWSA(self.calc.wfs.kd, pd,
                            timer=self.timer, txt=self.fd)

        pb = ProgressBar(self.fd)
        for kn, (s, ik, n1, n2) in pb.enumerate(self.mysKn1n2):
            Kstar_k = PWSA.unfold_ibz_kpoint(ik)
            for K_k in PWSA.group_kpoints(Kstar_k):
                # Let the first kpoint of the group represent
                # the rest of the kpoints
                K1 = K_k[0]
                # In this way wavefunctions are only loaded into
                # memory for this particular set of kpoints
                kptpair = self.get_kpoint_pair(pd, s, K1, n1, n2, m1, m2)
                kpt1 = kptpair.get_k1()  # kpt1 = k

                if kpt1.s not in spins:
                    continue
                kpt2 = kptpair.get_k2()  # kpt2 = k + q

                if unsymmetrized:
                    # Number of times kpoints are mapped into themselves
                    weight = np.sqrt(PWSA.how_many_symmetries() / len(K_k))

                # Use kpt2 to compute intraband transitions
                # These conditions are sufficient to make sure
                # that it still works in parallel
                if kpt1.n1 == 0 and self.blockcomm.rank == 0 and \
                   optical_limit and intraband:
                    assert self.nocc2 <= kpt2.nb, \
                        print('Error: Too few unoccupied bands')
                    vel0_mv = self.intraband_pair_density(kpt2)
                    f_m = kpt2.f_n[kpt2.na - kpt2.n1:kpt2.nb - kpt2.n1]
                    with self.timer('intraband'):
                        if vel0_mv is not None:
                            if unsymmetrized:
                                yield (f_m, None, None,
                                       None, None, vel0_mv / weight)
                            else:
                                for K2 in K_k:
                                    vel_mv = PWSA.map_v(K1, K2, vel0_mv)
                                    yield (f_m, None, None,
                                           None, None, vel_mv)

                # Divide the occupied bands into chunks
                n_n = np.arange(n2 - n1)
                if use_more_memory == 0:
                    chunksize = 1
                else:
                    chunksize = np.ceil(len(n_n) *
                                        use_more_memory).astype(int)

                no_n = []
                for i in range(len(n_n) // chunksize):
                    i1 = i * chunksize
                    i2 = min((i + 1) * chunksize, len(n_n))
                    no_n.append(n_n[i1:i2])

                # n runs over occupied bands
                for n_n in no_n:  # n_n is a list of occupied band indices
                    # m over unoccupied bands
                    m_m = np.arange(0, kpt2.n2 - kpt2.n1)
                    deps_nm = kptpair.get_transition_energies(n_n, m_m)
                    df_nm = kptpair.get_occupation_differences(n_n, m_m)

                    # This is not quite right for
                    # degenerate partially occupied
                    # bands, but good enough for now:
                    df_nm[df_nm <= 1e-20] = 0.0

                    # Get pair density for representative kpoint
                    ol = optical_limit
                    n0_nmG, n0_nmv, _ = self.get_pair_density(pd, kptpair,
                                                              n_n, m_m,
                                                              optical_limit=ol,
                                                              intraband=False,
                                                              Q_aGii=Q_aGii)

                    n0_nmG[deps_nm >= 0.0] = 0.0
                    if optical_limit:
                        n0_nmv[deps_nm >= 0.0] = 0.0

                    # Reshape nm -> m
                    nG = pd.ngmax
                    deps_m = deps_nm.reshape(-1)
                    df_m = df_nm.reshape(-1)
                    n0_mG = n0_nmG.reshape((-1, nG))
                    if optical_limit:
                        n0_mv = n0_nmv.reshape((-1, 3))

                    if unsymmetrized:
                        if optical_limit:
                            yield (None, df_m, deps_m,
                                   n0_mG / weight, n0_mv / weight, None)
                        else:
                            yield (None, df_m, deps_m,
                                   n0_mG / weight, None, None)
                        continue

                    # Collect pair densities in a single array
                    # and return them
                    nm = n0_mG.shape[0]
                    nG = n0_mG.shape[1]
                    nk = len(K_k)

                    n_MG = np.empty((nm * nk, nG), complex)
                    if optical_limit:
                        n_Mv = np.empty((nm * nk, 3), complex)
                    deps_M = np.tile(deps_m, nk)
                    df_M = np.tile(df_m, nk)

                    for i, K2 in enumerate(K_k):
                        i1 = i * nm
                        i2 = (i + 1) * nm
                        n_mG = PWSA.map_G(K1, K2, n0_mG)

                        if optical_limit:
                            n_mv = PWSA.map_v(K1, K2, n0_mv)
                            n_mG[:, 0] = n_mv[:, 0]
                            n_Mv[i1:i2, :] = n_mv

                        n_MG[i1:i2, :] = n_mG

                    if optical_limit:
                        yield (None, df_M, deps_M, n_MG, n_Mv, None)
                    else:
                        yield (None, df_M, deps_M, n_MG, None, None)

        pb.finish()

    @timer('Get kpoint pair')
    def get_kpoint_pair(self, pd, s, Kork_c, n1, n2, m1, m2,
                        load_wfs=True, block=False):
        # wfs = self.calc.wfs
        # bzk_kc = wfs.kd.bzk_kc

        if isinstance(Kork_c, int):
            # If k_c is an integer then it refers to
            # the index of the kpoint in the BZ
            k_c = self.calc.wfs.kd.bzk_kc[Kork_c]
        else:
            k_c = Kork_c

        q_c = pd.kd.bzk_kc[0]
        with self.timer('get k-points'):
            kpt1 = self.get_k_point(s, k_c, n1, n2, load_wfs=load_wfs)
            # K2 = wfs.kd.find_k_plus_q(q_c, [kpt1.K])[0]
            if self.response in ['+-', '-+']:
                s2 = 1 - s
            else:
                s2 = s
            kpt2 = self.get_k_point(s2, k_c + q_c, m1, m2,
                                    load_wfs=load_wfs, block=block)

        with self.timer('fft indices'):
            Q_G = self.get_fft_indices(kpt1.K, kpt2.K, q_c, pd,
                                       kpt1.shift_c - kpt2.shift_c)

        return KPointPair(kpt1, kpt2, Q_G)

    @timer('get_pair_density')
    def get_pair_density(self, pd, kptpair, n_n, m_m,
                         optical_limit=False, intraband=False,
                         Q_aGii=None, block=False, direction=2,
                         extend_head=True):
        """Get pair density for a kpoint pair."""
        ol = optical_limit = np.allclose(pd.kd.bzk_kc[0], 0.0) and \
            self.response == 'density'
        eh = extend_head
        cpd = self.calculate_pair_densities  # General pair densities
        opd = self.optical_pair_density  # Interband pair densities / q

        if Q_aGii is None:
            Q_aGii = self.initialize_paw_corrections(pd)

        kpt1 = kptpair.kpt1
        kpt2 = kptpair.kpt2
        Q_G = kptpair.Q_G  # Fourier components of kpoint pair
        nG = len(Q_G)

        if extend_head:
            n_nmG = np.zeros((len(n_n), len(m_m), nG + 2 * ol), pd.dtype)
        else:
            n_nmG = np.zeros((len(n_n), len(m_m), nG), pd.dtype)

        for j, n in enumerate(n_n):
            Q_G = kptpair.Q_G
            with self.timer('conj'):
                ut1cc_R = kpt1.ut_nR[n - kpt1.na].conj()
            with self.timer('paw'):
                C1_aGi = [np.dot(Q_Gii, P1_ni[n - kpt1.na].conj())
                          for Q_Gii, P1_ni in zip(Q_aGii, kpt1.P_ani)]
                n_nmG[j, :, 2 * ol * eh:] = cpd(ut1cc_R, C1_aGi, kpt2, pd, Q_G,
                                                block=block)
            if optical_limit:
                if extend_head:
                    n_nmG[j, :, 0:3] = opd(n, m_m, kpt1, kpt2,
                                           block=block)
                else:
                    n_nmG[j, :, 0] = opd(n, m_m, kpt1, kpt2,
                                         block=block)[:, direction]
        return n_nmG

    @timer('get_pair_momentum')
    def get_pair_momentum(self, pd, kptpair, n_n, m_m, Q_avGii=None):
        """Calculate matrix elements of the momentum operator.

        Calculates::

          n_{nm\mathrm{k}}\int_{\Omega_{\mathrm{cell}}}\mathrm{d}\mathbf{r}
          \psi_{n\mathrm{k}}^*(\mathbf{r})
          e^{-i\,(\mathrm{q} + \mathrm{G})\cdot\mathbf{r}}
          \nabla\psi_{m\mathrm{k} + \mathrm{q}}(\mathbf{r})

        pd: PlaneWaveDescriptor
            Plane wave descriptor of a single q_c.
        kptpair: KPointPair
            KpointPair object containing the two kpoints.
        n_n: list
            List of left-band indices (n).
        m_m:
            List of right-band indices (m).
        """
        wfs = self.calc.wfs

        kpt1 = kptpair.kpt1
        kpt2 = kptpair.kpt2
        Q_G = kptpair.Q_G  # Fourier components of kpoint pair

        # For the same band we
        kd = wfs.kd
        gd = wfs.gd
        k_c = kd.bzk_kc[kpt1.K] + kpt1.shift_c
        k_v = 2 * np.pi * np.dot(k_c, np.linalg.inv(gd.cell_cv).T)

        # Calculate k + G
        G_Gv = pd.get_reciprocal_vectors(add_q=True)
        kqG_Gv = k_v[np.newaxis] + G_Gv

        # Pair velocities
        n_nmvG = pd.zeros((len(n_n), len(m_m), 3))

        # Calculate derivatives of left-wavefunction
        # (there will typically be fewer of these)
        ut_nvR = self.make_derivative(kpt1.s, kpt1.K, kpt1.n1, kpt1.n2)

        # PAW-corrections
        if Q_avGii is None:
            Q_avGii = self.initialize_paw_nabla_corrections(pd)

        # Iterate over occupied bands
        for j, n in enumerate(n_n):
            ut1cc_R = kpt1.ut_nR[n].conj()

            n_mG = self.calculate_pair_densities(ut1cc_R,
                                                 [], kpt2,
                                                 pd, Q_G)

            n_nmvG[j] = 1j * kqG_Gv.T[np.newaxis] * n_mG[:, np.newaxis]

            # Treat each cartesian component at a time
            for v in range(3):
                # Minus from integration by parts
                utvcc_R = -ut_nvR[n, v].conj()
                Cv1_aGi = [np.dot(P1_ni[n].conj(), Q_vGii[v])
                           for Q_vGii, P1_ni in zip(Q_avGii, kpt1.P_ani)]

                nv_mG = self.calculate_pair_densities(utvcc_R,
                                                      Cv1_aGi, kpt2,
                                                      pd, Q_G)

                n_nmvG[j, :, v] += nv_mG

        # We want the momentum operator
        n_nmvG *= -1j

        return n_nmvG

    @timer('Calculate pair-densities')
    def calculate_pair_densities(self, ut1cc_R, C1_aGi, kpt2, pd, Q_G,
                                 block=True):
        """Calculate FFT of pair-densities and add PAW corrections.

        ut1cc_R: 3-d complex ndarray
            Complex conjugate of the periodic part of the left hand side
            wave function.
        C1_aGi: list of ndarrays
            PAW corrections for all atoms.
        kpt2: KPoint object
            Right hand side k-point object.
        pd: PWDescriptor
            Plane-wave descriptor for for q=k2-k1.
        Q_G: 1-d int ndarray
            Mapping from flattened 3-d FFT grid to 0.5(G+q)^2<ecut sphere.
        """

        dv = pd.gd.dv
        n_mG = pd.empty(kpt2.blocksize)
        myblocksize = kpt2.nb - kpt2.na

        for ut_R, n_G in zip(kpt2.ut_nR, n_mG):
            n_R = ut1cc_R * ut_R
            with self.timer('fft'):
                n_G[:] = pd.fft(n_R, 0, Q_G) * dv

        # PAW corrections:
        with self.timer('gemm'):
            for C1_Gi, P2_mi in zip(C1_aGi, kpt2.P_ani):
                gemm(1.0, C1_Gi, P2_mi, 1.0, n_mG[:myblocksize], 't')

        if not block or self.blockcomm.size == 1:
            return n_mG
        else:
            n_MG = pd.empty(kpt2.blocksize * self.blockcomm.size)
            self.blockcomm.all_gather(n_mG, n_MG)
            return n_MG[:kpt2.n2 - kpt2.n1]

    @timer('Optical limit')
    def optical_pair_velocity(self, n, m_m, kpt1, kpt2, block=False):
        if self.ut_sKnvR is None or kpt1.K not in self.ut_sKnvR[kpt1.s]:
            self.ut_sKnvR = self.calculate_derivatives(kpt1)

        kd = self.calc.wfs.kd
        gd = self.calc.wfs.gd
        k_c = kd.bzk_kc[kpt1.K] + kpt1.shift_c
        k_v = 2 * np.pi * np.dot(k_c, np.linalg.inv(gd.cell_cv).T)

        ut_vR = self.ut_sKnvR[kpt1.s][kpt1.K][n - kpt1.n1]
        atomdata_a = self.calc.wfs.setups
        C_avi = [np.dot(atomdata.nabla_iiv.T, P_ni[n - kpt1.na])
                 for atomdata, P_ni in zip(atomdata_a, kpt1.P_ani)]

        blockbands = kpt2.nb - kpt2.na
        n0_mv = np.empty((kpt2.blocksize, 3), dtype=complex)
        nt_m = np.empty(kpt2.blocksize, dtype=complex)
        n0_mv[:blockbands] = -self.calc.wfs.gd.integrate(ut_vR,
                                                         kpt2.ut_nR).T
        nt_m[:blockbands] = self.calc.wfs.gd.integrate(kpt1.ut_nR[n - kpt1.na],
                                                       kpt2.ut_nR)

        n0_mv[:blockbands] += (1j * nt_m[:blockbands, np.newaxis] *
                               k_v[np.newaxis, :])

        for C_vi, P_mi in zip(C_avi, kpt2.P_ani):
            gemm(1.0, C_vi, P_mi, 1.0, n0_mv[:blockbands], 'c')

        if block and self.blockcomm.size > 1:
            n0_Mv = np.empty((kpt2.blocksize * self.blockcomm.size, 3),
                             dtype=complex)
            self.blockcomm.all_gather(n0_mv, n0_Mv)
            n0_mv = n0_Mv[:kpt2.n2 - kpt2.n1]

        return -1j * n0_mv

    def optical_pair_density(self, n, m_m, kpt1, kpt2,
                             block=False):
        # Relative threshold for perturbation theory
        threshold = self.threshold

        eps1 = kpt1.eps_n[n - kpt1.n1]
        deps_m = (eps1 - kpt2.eps_n)[m_m - kpt2.n1]
        n0_mv = self.optical_pair_velocity(n, m_m, kpt1, kpt2,
                                           block=block)

        deps_m = deps_m.copy()
        deps_m[deps_m == 0.0] = np.inf

        smallness_mv = np.abs(-1e-3 * n0_mv / deps_m[:, np.newaxis])
        inds_mv = (np.logical_and(np.inf > smallness_mv,
                                  smallness_mv > threshold))
        n0_mv *= - 1 / deps_m[:, np.newaxis]
        n0_mv[inds_mv] = 0

        return n0_mv

    @timer('Intraband')
    def intraband_pair_density(self, kpt, n_n=None,
                               only_partially_occupied=False):
        """Calculate intraband matrix elements of nabla"""
        # Bands and check for block parallelization
        na, nb, n1 = kpt.na, kpt.nb, kpt.n1
        vel_nv = np.zeros((nb - na, 3), dtype=complex)
        if n_n is None:
            n_n = np.arange(na, nb)
        assert np.max(n_n) < nb, print('This is too many bands')
        assert np.min(n_n) >= na, print('This is too few bands')

        # Load kpoints
        kd = self.calc.wfs.kd
        gd = self.calc.wfs.gd
        k_c = kd.bzk_kc[kpt.K] + kpt.shift_c
        k_v = 2 * np.pi * np.dot(k_c, np.linalg.inv(gd.cell_cv).T)
        atomdata_a = self.calc.wfs.setups
        f_n = kpt.f_n

        # No carriers when T=0
        width = self.calc.occupations.width

        # Only works with Fermi-Dirac distribution
        assert isinstance(self.calc.occupations, FermiDirac)
        if width > 1e-15:
            dfde_n = -1 / width * (f_n - f_n**2.0)  # Analytical derivative
            partocc_n = np.abs(dfde_n) > 1e-5  # Is part. occupied?
        else:
            # Just include all bands to be sure
            partocc_n = np.ones(len(f_n), dtype=bool)

        if only_partially_occupied and not partocc_n.any():
            return None

        if only_partially_occupied:
            # Check for block par. consistency
            assert (partocc_n < nb).all(), \
                print('Include more unoccupied bands ', +
                      'or less block parr.', file=self.fd)

        # Break bands into degenerate chunks
        degchunks_cn = []  # indexing c as chunk number
        for n in n_n:
            inds_n = np.nonzero(np.abs(kpt.eps_n[n - n1] -
                                       kpt.eps_n) < 1e-5)[0] + n1

            # Has this chunk already been computed?
            oldchunk = any([n in chunk for chunk in degchunks_cn])
            if not oldchunk and \
               (partocc_n[n - n1] or not only_partially_occupied):
                assert all([ind in n_n for ind in inds_n]), \
                    print('\nYou are cutting over a degenerate band ' +
                          'using block parallelization.',
                          inds_n, n_n, file=self.fd)
                degchunks_cn.append((inds_n))

        # Calculate matrix elements by diagonalizing each block
        for ind_n in degchunks_cn:
            deg = len(ind_n)
            ut_nvR = self.calc.wfs.gd.zeros((deg, 3), complex)
            vel_nnv = np.zeros((deg, deg, 3), dtype=complex)
            # States are included starting from kpt.na
            ut_nR = kpt.ut_nR[ind_n - na]

            # Get derivatives
            for ind, ut_vR in zip(ind_n, ut_nvR):
                ut_vR[:] = self.make_derivative(kpt.s, kpt.K,
                                                ind, ind + 1)[0]

            # Treat the whole degenerate chunk
            for n in range(deg):
                ut_vR = ut_nvR[n]
                C_avi = [np.dot(atomdata.nabla_iiv.T, P_ni[ind_n[n] - na])
                         for atomdata, P_ni in zip(atomdata_a, kpt.P_ani)]

                nabla0_nv = -self.calc.wfs.gd.integrate(ut_vR, ut_nR).T
                nt_n = self.calc.wfs.gd.integrate(ut_nR[n], ut_nR)
                nabla0_nv += 1j * nt_n[:, np.newaxis] * k_v[np.newaxis, :]

                for C_vi, P_ni in zip(C_avi, kpt.P_ani):
                    gemm(1.0, C_vi, P_ni[ind_n - na], 1.0, nabla0_nv, 'c')

                vel_nnv[n] = -1j * nabla0_nv

            for iv in range(3):
                vel, _ = np.linalg.eig(vel_nnv[..., iv])
                vel_nv[ind_n - na, iv] = vel  # Use eigenvalues

        return vel_nv[n_n - na]

    def get_fft_indices(self, K1, K2, q_c, pd, shift0_c):
        """Get indices for G-vectors inside cutoff sphere."""
        kd = self.calc.wfs.kd
        N_G = pd.Q_qG[0]
        shift_c = (shift0_c +
                   (q_c - kd.bzk_kc[K2] + kd.bzk_kc[K1]).round().astype(int))
        if shift_c.any():
            n_cG = np.unravel_index(N_G, pd.gd.N_c)
            n_cG = [n_G + shift for n_G, shift in zip(n_cG, shift_c)]
            N_G = np.ravel_multi_index(n_cG, pd.gd.N_c, 'wrap')
        return N_G

    def construct_symmetry_operators(self, K, k_c=None):
        """Construct symmetry operators for wave function and PAW projections.

        We want to transform a k-point in the irreducible part of the BZ to
        the corresponding k-point with index K.

        Returns U_cc, T, a_a, U_aii, shift_c and time_reversal, where:

        * U_cc is a rotation matrix.
        * T() is a function that transforms the periodic part of the wave
          function.
        * a_a is a list of symmetry related atom indices
        * U_aii is a list of rotation matrices for the PAW projections
        * shift_c is three integers: see code below.
        * time_reversal is a flag - if True, projections should be complex
          conjugated.

        See the get_k_point() method for how to use these tuples.
        """

        wfs = self.calc.wfs
        kd = wfs.kd

        s = kd.sym_k[K]
        U_cc = kd.symmetry.op_scc[s]
        time_reversal = kd.time_reversal_k[K]
        ik = kd.bz2ibz_k[K]
        if k_c is None:
            k_c = kd.bzk_kc[K]
        ik_c = kd.ibzk_kc[ik]

        sign = 1 - 2 * time_reversal
        shift_c = np.dot(U_cc, ik_c) - k_c * sign

        try:
            assert np.allclose(shift_c.round(), shift_c)
        except AssertionError:
            print('shift_c ' + str(shift_c), file=self.fd)
            print('k_c ' + str(k_c), file=self.fd)
            print('kd.bzk_kc[K] ' + str(kd.bzk_kc[K]), file=self.fd)
            print('ik_c ' + str(ik_c), file=self.fd)
            print('U_cc ' + str(U_cc), file=self.fd)
            print('sign ' + str(sign), file=self.fd)
            raise AssertionError

        shift_c = shift_c.round().astype(int)

        if (U_cc == np.eye(3)).all():
            T = lambda f_R: f_R
        else:
            N_c = self.calc.wfs.gd.N_c
            i_cr = np.dot(U_cc.T, np.indices(N_c).reshape((3, -1)))
            i = np.ravel_multi_index(i_cr, N_c, 'wrap')
            T = lambda f_R: f_R.ravel()[i].reshape(N_c)

        if time_reversal:
            T0 = T
            T = lambda f_R: T0(f_R).conj()
            shift_c *= -1

        a_a = []
        U_aii = []
        for a, id in enumerate(self.calc.wfs.setups.id_a):
            b = kd.symmetry.a_sa[s, a]
            S_c = np.dot(self.spos_ac[a], U_cc) - self.spos_ac[b]
            x = np.exp(2j * pi * np.dot(ik_c, S_c))
            U_ii = wfs.setups[a].R_sii[s].T * x
            a_a.append(b)
            U_aii.append(U_ii)

        return U_cc, T, a_a, U_aii, shift_c, time_reversal

    @timer('Initialize PAW corrections')
    def initialize_paw_corrections(self, pd, soft=False):
        wfs = self.calc.wfs
        q_v = pd.K_qv[0]
        optical_limit = np.allclose(q_v, 0) and self.response == 'density'

        G_Gv = pd.get_reciprocal_vectors()
        if optical_limit:
            G_Gv[0] = 1

        pos_av = np.dot(self.spos_ac, pd.gd.cell_cv)

        # Collect integrals for all species:
        Q_xGii = {}
        for id, atomdata in wfs.setups.setups.items():
            if soft:
                ghat = PWLFC([atomdata.ghat_l], pd)
                ghat.set_positions(np.zeros((1, 3)))
                Q_LG = ghat.expand().T
                if atomdata.Delta_iiL is None:
                    ni = atomdata.ni
                    Q_Gii = np.zeros((Q_LG.shape[1], ni, ni))
                else:
                    Q_Gii = np.dot(atomdata.Delta_iiL, Q_LG).T
            else:
                Q_Gii = two_phi_planewave_integrals(G_Gv, atomdata)
                ni = atomdata.ni
                Q_Gii.shape = (-1, ni, ni)

            Q_xGii[id] = Q_Gii

        Q_aGii = []
        for a, atomdata in enumerate(wfs.setups):
            id = wfs.setups.id_a[a]
            Q_Gii = Q_xGii[id]
            x_G = np.exp(-1j * np.dot(G_Gv, pos_av[a]))
            Q_aGii.append(x_G[:, np.newaxis, np.newaxis] * Q_Gii)
            if optical_limit:
                Q_aGii[a][0] = atomdata.dO_ii

        return Q_aGii

    @timer('Initialize PAW corrections')
    def initialize_paw_nabla_corrections(self, pd, soft=False):
        print('Initializing nabla PAW Corrections', file=self.fd)
        wfs = self.calc.wfs
        G_Gv = pd.get_reciprocal_vectors()
        pos_av = np.dot(self.spos_ac, pd.gd.cell_cv)

        # Collect integrals for all species:
        Q_xvGii = {}
        for id, atomdata in wfs.setups.setups.items():
            if soft:
                raise NotImplementedError
            else:
                Q_vGii = two_phi_nabla_planewave_integrals(G_Gv, atomdata)
                ni = atomdata.ni
                Q_vGii.shape = (3, -1, ni, ni)

            Q_xvGii[id] = Q_vGii

        Q_avGii = []
        for a, atomdata in enumerate(wfs.setups):
            id = wfs.setups.id_a[a]
            Q_vGii = Q_xvGii[id]
            x_G = np.exp(-1j * np.dot(G_Gv, pos_av[a]))
            Q_avGii.append(x_G[np.newaxis, :, np.newaxis, np.newaxis] * Q_vGii)

        return Q_avGii

    def calculate_derivatives(self, kpt):
        ut_sKnvR = [{}, {}]
        ut_nvR = self.make_derivative(kpt.s, kpt.K, kpt.n1, kpt.n2)
        ut_sKnvR[kpt.s][kpt.K] = ut_nvR

        return ut_sKnvR

    @timer('Derivatives')
    def make_derivative(self, s, K, n1, n2):
        wfs = self.calc.wfs
        if self.real_space_derivatives:
            grad_v = [Gradient(wfs.gd, v, 1.0, 4, complex).apply
                      for v in range(3)]

        U_cc, T, a_a, U_aii, shift_c, time_reversal = \
            self.construct_symmetry_operators(K)
        A_cv = wfs.gd.cell_cv
        M_vv = np.dot(np.dot(A_cv.T, U_cc.T), np.linalg.inv(A_cv).T)
        ik = wfs.kd.bz2ibz_k[K]
        kpt = wfs.kpt_u[s * wfs.kd.nibzkpts + ik]
        psit_nG = kpt.psit_nG
        iG_Gv = 1j * wfs.pd.get_reciprocal_vectors(q=ik, add_q=False)
        ut_nvR = wfs.gd.zeros((n2 - n1, 3), complex)
        for n in range(n1, n2):
            for v in range(3):
                if self.real_space_derivatives:
                    ut_R = T(wfs.pd.ifft(psit_nG[n], ik))
                    grad_v[v](ut_R, ut_nvR[n - n1, v],
                              np.ones((3, 2), complex))
                else:
                    ut_R = T(wfs.pd.ifft(iG_Gv[:, v] * psit_nG[n], ik))
                    for v2 in range(3):
                        ut_nvR[n - n1, v2] += ut_R * M_vv[v, v2]

        return ut_nvR



from __future__ import print_function
import numpy as np

from ase.build import molecule

from gpaw import GPAW
from gpaw.lrtddft2 import LrTDDFT2
from gpaw.test import equal

name = 'H2O-lcao'
atoms = molecule('H2O')
atoms.center(vacuum=4)

# Ground state
calc = GPAW(h=0.4, mode='lcao', basis='dzp', txt='%s-gs.out' % name,
            poissonsolver={'name': 'fd'},
            nbands=8, xc='LDA')
atoms.set_calculator(calc)
atoms.get_potential_energy()
calc.write('%s.gpw' % name, mode='all')

# LrTDDFT2
calc = GPAW('%s.gpw' % name, txt='%s-lr.out' % name)
lr = LrTDDFT2(name, calc, fxc='LDA')
lr.calculate()
results = lr.get_transitions()[0:2]

if 0:
    np.set_printoptions(precision=10)
    refstr = repr(results)
    refstr = refstr.replace('array', 'np.array')
    # Format a pep-compatible output
    refstr = ' '.join(refstr.split())
    refstr = refstr.replace('[ ', '[')
    refstr = refstr.replace(', np', ',\n       np')
    refstr = refstr.replace(', ', ',\n                 ')
    print('ref = %s' % refstr)

ref = (np.array([6.0832418565,
                 8.8741524165,
                 13.5935582927,
                 14.291759976,
                 15.9923574087,
                 16.9926770576,
                 17.6505087373,
                 17.6924715713,
                 24.092978682,
                 25.0027646421,
                 25.6208743025,
                 26.9649406073,
                 29.5294888977,
                 29.8439220428]),
       np.array([3.6910848669e-02,
                 5.4510953134e-24,
                 3.0995301001e-01,
                 2.1335662795e-02,
                 7.8725024176e-22,
                 6.3875174151e-02,
                 1.0288169027e-01,
                 1.7216407337e-01,
                 2.8901383040e-02,
                 3.9354412451e-01,
                 2.1917641289e-02,
                 6.7746017219e-01,
                 7.9561718249e-02,
                 1.0625830362e-02]))

tol = 1e-3
for r0, r1 in zip(results, ref):
    equal(r0, r1, tol)



"""Run AGTS tests.

Initial setup::

    mkdir agts
    cd agts
    git clone http://gitlab.com/ase/ase.git
    git clone http://gitlab.com/gpaw/gpaw.git

Crontab::

    WEB_PAGE_FOLDER=...
    AGTS=...
    CMD="python $AGTS/gpaw/utilities/agts.py"
    10 20 * * 5 cd $AGTS; $CMD run > agts-run.log
    10 20 * * 1 cd $AGTS; $CMD summary > agts-summary.log

"""

import datetime
import functools
import json
import os
import subprocess
from pathlib import Path


shell = functools.partial(subprocess.check_call, shell=True)


def agts(cmd):
    from myqueue.task import taskstates
    from myqueue.tasks import Tasks, Selection

    allofthem = Selection(None, '', taskstates, [Path('.').absolute()], True)
    with Tasks(verbosity=-1) as t:
        tasks = t.list(allofthem, '')

    print(len(tasks))

    if cmd == 'run':
        if tasks:
            raise ValueError('Not ready!')

        arch = 'linux-x86_64-broadwell-el7-3.6'
        path = '/home/niflheim2/jensj/agts'
        pp = ('{path}/ase:{path}/gpaw:{path}/gpaw/build/lib.{arch}:$PYTHONPATH'
              .format(path=path, arch=arch))

        shell('cd ase;'
              'git pull')
        shell('cd gpaw;'
              'git clean -fdx;'
              'git pull;'
              '. doc/platforms/Linux/Niflheim/compile.sh;'
              'cd doc;'
              'PYTHONPATH={pp} make'
              .format(pp=pp))
        shell('PYTHONPATH={pp} mq workflow -p agts.py gpaw -T'
              .format(pp=pp))

    elif cmd == 'summary':
        write_results(tasks)

        for task in tasks:
            if task.state in {'running', 'queued'}:
                raise RuntimeError('Not done!')

        for task in tasks:
            if task.state in {'FAILED', 'CANCELED', 'TIMEOUT'}:
                send_email(tasks)
                return

        shell('mq remove --state d --recursive .')

        collect_files_for_web_page()
    else:
        1 / 0


def send_email(tasks):
    import smtplib
    from email.message import EmailMessage

    txt = 'Hi!\n\n'
    for task in tasks:
        if task.state in {'FAILED', 'CANCELED', 'TIMEOUT', 'MEMORY'}:
            id, dir, name, res, age, status, t, err = task.words()
            txt += ('test: {}/{}@{}: {}\ntime: {}\nerror: {}\n\n'
                    .format(dir.split('agts/gpaw')[1],
                            name,
                            res[:-1],
                            status,
                            t,
                            err))
    txt += 'Best regards,\nNiflheim\n'

    msg = EmailMessage()
    msg.set_content(txt)
    msg['Subject'] = 'Failing Niflheim-tests!'
    msg['From'] = 'agts@niflheim.dtu.dk'
    msg['To'] = 'jjmo@dtu.dk'
    s = smtplib.SMTP('smtp.ait.dtu.dk')
    s.send_message(msg)
    s.quit()


def find_created_files():
    names = set()
    for path in Path().glob('**/*.py'):
        if path.parts[0] == 'build':
            continue
        line1 = path.read_text().split('\n', 1)[0]
        if not line1.startswith('# Creates:'):
            continue
        for name in line1.split(':')[1].split(','):
            name = name.strip()
            if name in names:
                raise RuntimeError(
                    'The name {!r} is used in more than one place!'
                    .format(name))
            names.add(name)
            yield path.with_name(name)


def collect_files_for_web_page():
    os.chdir('gpaw/doc')
    folder = Path('agts-files')
    if not folder.is_dir():
        folder.mkdir()
    for path in find_created_files():
        print(path)
        (folder / path.name).write_bytes(path.read_bytes())
    # os.environ['WEB_PAGE_FOLDER']


def write_results(tasks):
    now = datetime.datetime.now()
    name = Path('{}-{:02}-{:02}.json'.format(now.year, now.month, now.day))
    text = json.dumps({'tasks': [task.todict() for task in tasks]},
                      indent=2)
    name.write_text(text)


def compare(f1, f2):
    dicts = []
    for f in [f1, f2]:
        d = json.loads(Path(f).read_text())
        dct = {}
        for task in d['tasks']:
            name = task['folder'] + '/' + task['cmd']['cmd']
            if task['cmd']['args']:
                name += '_' + '+'.join(task['cmd']['args'])
            dct[name] = task['tstop'] - task['trunning']
        dicts.append(dct)
    d1, d2 = dicts
    data = []
    for name1, t1 in d1.items():
        if name1 in d2:
            data.append((d2[name1] - t1, t1, name1))
    for dt, t, name in sorted(data):
        name = name.split('agts/gpaw/')[1]
        print('{dt:+10.1f} {dt / t * 100:+6.1f} {t:10.1f} {name}'
              .format(**locals()))


if __name__ == '__main__':
    import sys
    cmd = sys.argv[1]
    if cmd == 'compare':
        compare(*sys.argv[2:])
    else:
        agts(cmd)



import numpy as np
from ase.units import Bohr
from gpaw.poisson import FDPoissonSolver
from gpaw.utilities.gauss import Gaussian
from gpaw.utilities.timing import nulltimer

from ase.utils.timing import timer

from gpaw.utilities.extend_grid import extended_grid_descriptor, \
    extend_array, deextend_array


class ExtendedPoissonSolver(FDPoissonSolver):
    """ExtendedPoissonSolver

    Parameter syntax:

    moment_corrections = [{'moms': moms_list1, 'center': center1},
                          {'moms': moms_list2, 'center': center2},
                          ...]
    Here moms_listX is list of integers of multipole moments to be corrected
    at centerX.

    extended = {'gpts': gpts, 'useprev': useprev}
    Here gpts is number of grid points in the **coarse** grid corresponding
    to the larger grid used for PoissonSolver (the Poisson equation
    is solved on fine grid as usual, but the gpts is given in coarse grid
    units for convenience), and useprev is boolean determining whether previous
    solution of the PoissonSolver instance is used as an initial guess
    for the next solve() call.

    Important: provide timer for PoissonSolver to analyze the cost of
    the multipole moment corrections and grid extension to your system!

    """
    # TODO: enable 'comm' parameter for 'extended' dictionary. This would
    # allow to use the whole mpi.world for PoissonSolver.
    # Currently, Poisson equation calculation is duplicated over, e.g.,
    # band and kpt communicators.

    def __init__(self, nn=3, relax='J', eps=2e-10, maxiter=1000,
                 moment_corrections=None,
                 extended=None,
                 timer=nulltimer):

        FDPoissonSolver.__init__(self, nn=nn, relax=relax,
                                 eps=eps, maxiter=maxiter,
                                 remove_moment=None)

        self.timer = timer

        if moment_corrections is None:
            self.moment_corrections = None
        elif isinstance(moment_corrections, int):
            self.moment_corrections = [{'moms': range(moment_corrections),
                                        'center': None}]
        else:
            self.moment_corrections = moment_corrections

        self.is_extended = False
        # Broadcast over band, kpt, etc. communicators required?
        self.requires_broadcast = False
        if extended is not None:
            self.is_extended = True
            self.extended = extended
            assert 'gpts' in extended.keys(), 'gpts parameter is missing'
            self.extended['gpts'] = np.array(self.extended['gpts'])
            # Multiply gpts by 2 to get gpts on fine grid
            self.extended['finegpts'] = self.extended['gpts'] * 2
            assert 'useprev' in extended.keys(), 'useprev parameter is missing'
            if self.extended.get('comm') is not None:
                self.requires_broadcast = True

    def set_grid_descriptor(self, gd):
        if self.is_extended:
            self.gd_original = gd
            assert np.all(self.gd_original.N_c <= self.extended['finegpts']), \
                'extended grid has to be larger than the original one'
            gd, _, _ = extended_grid_descriptor(
                gd,
                N_c=self.extended['finegpts'],
                extcomm=self.extended.get('comm'))
        FDPoissonSolver.set_grid_descriptor(self, gd)

    def get_description(self):
        description = FDPoissonSolver.get_description(self)

        lines = [description]

        if self.is_extended:
            lines.extend(['    Extended %d*%d*%d grid' %
                          tuple(self.gd.N_c)])
            lines.extend(['    Use previous is %s' %
                          self.extended['useprev']])

        if self.moment_corrections:
            lines.extend(['    %d moment corrections:' %
                          len(self.moment_corrections)])
            lines.extend(['      %s' %
                          ('[%s] %s' %
                           ('center' if mom['center'] is None
                            else (', '.join(['%.2f' % (x * Bohr)
                                             for x in mom['center']])),
                            mom['moms']))
                          for mom in self.moment_corrections])

        return '\n'.join(lines)

    @timer('Poisson initialize')
    def _init(self):
        if self._initialized:
            return
        FDPoissonSolver._init(self)

        if self.is_extended:
            if not self.gd.orthogonal or self.gd.pbc_c.any():
                raise NotImplementedError('Only orthogonal unit cells '
                                          'and non-periodic boundary '
                                          'conditions are tested')
            self.rho_g = self.gd.zeros()
            self.phi_g = self.gd.zeros()

        if self.moment_corrections is not None:
            if not self.gd.orthogonal or self.gd.pbc_c.any():
                raise NotImplementedError('Only orthogonal unit cells '
                                          'and non-periodic boundary '
                                          'conditions are tested')
            self.load_moment_corrections_gauss()

    @timer('Load moment corrections')
    def load_moment_corrections_gauss(self):
        if not hasattr(self, 'gauss_i'):
            self.gauss_i = []
            mask_ir = []
            r_ir = []
            self.mom_ij = []

            for rmom in self.moment_corrections:
                if rmom['center'] is None:
                    center = None
                else:
                    center = np.array(rmom['center'])
                mom_j = rmom['moms']
                gauss = Gaussian(self.gd, center=center)
                self.gauss_i.append(gauss)
                r_ir.append(gauss.r.ravel())
                mask_ir.append(self.gd.zeros(dtype=int).ravel())
                self.mom_ij.append(mom_j)

            r_ir = np.array(r_ir)
            mask_ir = np.array(mask_ir)

            Ni = r_ir.shape[0]
            Nr = r_ir.shape[1]

            for r in range(Nr):
                i = np.argmin(r_ir[:, r])
                mask_ir[i, r] = 1

            self.mask_ig = []
            for i in range(Ni):
                mask_r = mask_ir[i]
                mask_g = mask_r.reshape(self.gd.n_c)
                self.mask_ig.append(mask_g)

                # Uncomment this to see masks on grid
                # big_g = self.gd.collect(mask_g)
                # if self.gd.comm.rank == 0:
                #     big_g.dump('mask_%dg' % (i))

    def solve(self, phi, rho, charge=None, eps=None, maxcharge=1e-6,
              zero_initial_phi=False):
        self._init()
        if self.is_extended:
            self.rho_g[:] = 0
            if not self.extended['useprev']:
                self.phi_g[:] = 0

            self.timer.start('Extend array')
            extend_array(self.gd_original, self.gd, rho, self.rho_g)
            self.timer.stop('Extend array')

            retval = self._solve(self.phi_g, self.rho_g, charge,
                                 eps, maxcharge, zero_initial_phi)

            self.timer.start('Deextend array')
            deextend_array(self.gd_original, self.gd, phi, self.phi_g)
            self.timer.stop('Deextend array')

            return retval
        else:
            return self._solve(phi, rho, charge, eps, maxcharge,
                               zero_initial_phi)

    @timer('Solve')
    def _solve(self, phi, rho, charge=None, eps=None, maxcharge=1e-6,
               zero_initial_phi=False):
        self._init()
        if eps is None:
            eps = self.eps

        if self.moment_corrections:
            assert not self.gd.pbc_c.any()

            self.timer.start('Multipole moment corrections')

            rho_neutral = rho * 0.0
            phi_cor_k = []
            for gauss, mask_g, mom_j in zip(self.gauss_i, self.mask_ig,
                                            self.mom_ij):
                rho_masked = rho * mask_g
                for mom in mom_j:
                    phi_cor_k.append(gauss.remove_moment(rho_masked, mom))
                rho_neutral += rho_masked

            # Remove multipoles for better initial guess
            for phi_cor in phi_cor_k:
                phi -= phi_cor

            self.timer.stop('Multipole moment corrections')

            self.timer.start('Solve neutral')
            niter = self.solve_neutral(phi, rho_neutral, eps=eps)
            self.timer.stop('Solve neutral')

            self.timer.start('Multipole moment corrections')
            # correct error introduced by removing multipoles
            for phi_cor in phi_cor_k:
                phi += phi_cor
            self.timer.stop('Multipole moment corrections')

            return niter
        else:
            return FDPoissonSolver.solve(self, phi, rho, charge,
                                         eps, maxcharge,
                                         zero_initial_phi)

    def estimate_memory(self, mem):
        FDPoissonSolver.estimate_memory(self, mem)
        gdbytes = self.gd.bytecount()
        if self.is_extended:
            mem.subnode('extended arrays',
                        2 * gdbytes)
        if self.moment_corrections is not None:
            mem.subnode('moment_corrections masks',
                        len(self.moment_corrections) * gdbytes)

    def __repr__(self):
        template = 'ExtendedPoissonSolver(relax=\'%s\', nn=%s, eps=%e)'
        representation = template % (self.relax, repr(self.nn), self.eps)
        return representation



import numpy as np

from gpaw.test import equal
from gpaw.fftw import FFTPlan
from gpaw.grid_descriptor import GridDescriptor
from gpaw.spline import Spline
import gpaw.mpi as mpi
from gpaw.wavefunctions.pw import PWDescriptor, PWLFC


x = 2.0
rc = 3.5
r = np.linspace(0, rc, 100)

n = 40
a = 8.0
gd = GridDescriptor((n, n, n), (a, a, a), comm=mpi.serial_comm)

spos_ac = np.array([(0.15, 0.45, 0.95)])

pd = PWDescriptor(45, gd, complex)
pdr = PWDescriptor(45, gd)

print(FFTPlan)

for l in range(4):
    print(l)
    s = Spline(l, rc, 2 * x**1.5 / np.pi * np.exp(-x * r**2))

    lfc = PWLFC([[s]], pd)
    lfcr = PWLFC([[s]], pdr)

    c_axi = {0: np.zeros((1, 2 * l + 1), complex)}
    c_axi[0][0, 0] = 1.9
    cr_axi = {0: np.zeros((1, 2 * l + 1))}
    cr_axi[0][0, 0] = 1.9

    b = pd.zeros(1, dtype=complex)
    br = pdr.zeros(1)

    lfc.set_positions(spos_ac)
    lfc.add(b, c_axi)
    lfcr.set_positions(spos_ac)
    lfcr.add(br, cr_axi)

    a = pd.ifft(b)
    ar = pdr.ifft(br)
    equal(abs(a - ar).max(), 0, 1e-14)

    if l == 0:
        a = a[:, ::-1].copy()
        b0 = pd.fft(a)
        br0 = pdr.fft(a.real)

    lfc.integrate(b0, c_axi)
    lfcr.integrate(br0, cr_axi)

    assert abs(c_axi[0][0] - cr_axi[0][0]).max() < 1e-14

    c_axiv = {0: np.zeros((1, 2 * l + 1, 3), complex)}
    cr_axiv = {0: np.zeros((1, 2 * l + 1, 3))}
    lfc.derivative(b0, c_axiv)
    lfcr.derivative(br0, cr_axiv)
    assert abs(c_axiv[0][0] - cr_axiv[0][0]).max() < 1e-14



from ase import Atom, Atoms
from gpaw import GPAW, FermiDirac
from gpaw.test import equal

a = 4.0
n = 16
hydrogen = Atoms([Atom('H')], cell=(a, a, a), pbc=True)
hydrogen.center()
calc = GPAW(gpts=(n, n, n), nbands=1, convergence={'energy': 1e-5},
            occupations=FermiDirac(0.0))
hydrogen.set_calculator(calc)
e1 = hydrogen.get_potential_energy()
niter1 = calc.get_number_of_iterations()
hydrogen.set_initial_magnetic_moments([1.0])
e2 = hydrogen.get_potential_energy()
niter2 = calc.get_number_of_iterations()
de = e1 - e2
print(de)
equal(de, 0.7871, 1.e-4)

energy_tolerance = 0.0006
equal(e1, -0.499854, energy_tolerance)
equal(e2, -1.287, energy_tolerance)



from ase.build import bulk
from gpaw import GPAW, FermiDirac
from gpaw.mpi import serial_comm
from gpaw.test import equal
from gpaw.xc.rpa import RPACorrelation
from gpaw.xc.fxc import FXCCorrelation

a0 = 5.43
Ni = bulk('Ni', 'fcc')
Ni.set_initial_magnetic_moments([0.7])

calc = GPAW(mode='pw',
            kpts=(3, 3, 3),
            occupations=FermiDirac(0.001),
            setups={'Ni': '10'},
            communicator=serial_comm)
Ni.set_calculator(calc)
E = Ni.get_potential_energy()
calc.diagonalize_full_hamiltonian(nbands=50)

rpa = RPACorrelation(calc, nfrequencies=8, skip_gamma=True)
E_rpa = rpa.calculate(ecut=[50])

fxc = FXCCorrelation(calc, nlambda=16, nfrequencies=8, skip_gamma=True)
E_fxc = fxc.calculate(ecut=[50])

equal(E_rpa, -7.826, 0.01)
equal(E_fxc, -7.826, 0.01)



from gpaw.symmetry import map_k_points_fast, map_k_points
from ase.dft.kpoints import monkhorst_pack
from gpaw import GPAW
from ase.build import bulk
import numpy as np
from itertools import product


def test_mapping(bz2bz_ks, U_scc, bzk_kc, time_reversal=False):
    eps = 1e-7

    if time_reversal:
        U_scc = np.concatenate([U_scc, -U_scc])

    for k_c, bz2bz_s in zip(bzk_kc, bz2bz_ks):
        delta_sc = (np.dot(U_scc, k_c) -
                    bzk_kc[bz2bz_s, :])[bz2bz_s >= 0]
        delta_sc = np.abs(delta_sc - delta_sc.round())
        assert delta_sc.max() < eps


# Test kpoint mapping functionality of the gpaw.symmetry module
atoms = bulk('C')
calc = GPAW(mode='pw',
            kpts={'size': (5, 5, 5)},
            txt=None)
atoms.calc = calc
atoms.get_potential_energy()

U_scc = calc.wfs.kd.symmetry.op_scc
time_reversal = False

for gamma in [True, False]:
    for time_reversal in [True, False]:
        for i, j, k in product(*([range(1, 7)] * 3)):

            bzk_kc = monkhorst_pack((i, j, k))

            if gamma:
                offset = (((i + 1) % 2) / (2 * i),
                          ((j + 1) % 2) / (2 * j),
                          ((k + 1) % 2) / (2 * k))
                bzk_kc += offset

            bz2bz_ks = map_k_points(bzk_kc, U_scc,
                                    time_reversal,
                                    None, calc.wfs.kd.symmetry.tol)
            bz2bzfast_ks = map_k_points_fast(bzk_kc, U_scc,
                                             time_reversal,
                                             None,
                                             calc.wfs.kd.symmetry.tol)

            assert ((bz2bz_ks - bz2bzfast_ks)**2 < 1e-9).all()

            test_mapping(bz2bz_ks, U_scc, bzk_kc, time_reversal)
            test_mapping(bz2bzfast_ks, U_scc, bzk_kc, time_reversal)



from gpaw import GPAW, PW
from ase.lattice.cubic import BodyCenteredCubic
from gpaw.test import equal

QNA = {'alpha': 2.0,
       'name': 'QNA',
       'orbital_dependent': False,
       'parameters': {'Fe': (0.1485, 0.005)},
       'setup_name': 'PBE',
       'type': 'qna-gga'}

atoms = BodyCenteredCubic(symbol='Fe',
                          latticeconstant=2.854,
                          pbc=(1, 1, 1))

atoms.set_initial_magnetic_moments([2, 2])

calc = GPAW(mode=PW(400),
            kpts=(3, 3, 3),
            experimental={'niter_fixdensity': 2},
            xc=QNA,
            parallel={'domain': 1},
            txt='qna_spinpol.txt')

atoms.set_calculator(calc)
atoms.get_potential_energy()
magmoms = atoms.get_magnetic_moments()

tol = 0.003
equal(2.243, magmoms[0], tol)
equal(2.243, magmoms[1], tol)



"""Self-consistent SOC."""
from unittest import SkipTest

import numpy as np
from ase.build import mx2

from gpaw import GPAW
from gpaw.test import equal
from gpaw.spinorbit import get_spinorbit_eigenvalues
from gpaw.mpi import size


if size > 1:
    raise SkipTest()


a = mx2('MoS2')
a.center(vacuum=3, axis=2)

params = dict(mode='pw',
              kpts={'size': (3, 3, 1),
                    'gamma': True})

# Selfconsistent:
a.calc = GPAW(experimental={'magmoms': np.zeros((3, 3)),
                            'soc': True},
              convergence={'bands': 28},
              **params)
a.get_potential_energy()
E1 = a.calc.get_eigenvalues(kpt=2)

# Non-selfconsistent:
a.calc = GPAW(convergence={'bands': 14}, **params)
a.get_potential_energy()
E2 = get_spinorbit_eigenvalues(a.calc, bands=np.arange(14))[:, 2]


def test(E, hsplit, lsplit):
    print(E)
    h1, h2, l1, l2 = E[24:28]  # HOMO-1, HOMO, LUMO, LUMO+1
    print(h2 - h1)
    print(l2 - l1)
    assert abs(h2 - h1 - hsplit) < 0.01
    assert abs(l2 - l1 - lsplit) < 0.002


equal(E1[:28], E2, tolerance=1e-1)
test(E1, 0.15, 0.002)
test(E2, 0.15, 0.007)



from gpaw import GPAW, FermiDirac
from gpaw.mpi import world, size, rank
from gpaw.lrtddft2 import LrTDDFT2
from gpaw.lrtddft2.lr_communicators import LrCommunicators
from gpaw.test import equal
from ase.atoms import Atoms

debug = False
restart_file = 'Al2_gs.gpw'

d = 2.563
atoms = Atoms('Al2', positions=((0, 0, 0),
                                (0, 0, d)))
atoms.center(4.0)
calc = GPAW(h=0.24, eigensolver='cg', basis='dzp',
            occupations=FermiDirac(width=0.01),
            convergence={'eigenstates': 4.0e-5,
                         'density': 1.0e-2,
                         'bands': 'all'},
            nbands=20)
atoms.set_calculator(calc)
atoms.get_potential_energy()
calc.write(restart_file, mode='all')

# Try to run parallel over eh-pairs
if size % 2 == 0:
    eh_size = 2
    domain_size = size // eh_size
else:
    eh_size = 1
    domain_size = size

lr_comms = LrCommunicators(world, domain_size, eh_size)

calc = GPAW(restart_file,
            communicator=lr_comms.dd_comm)
de = 3.0
lr = LrTDDFT2('Al2_lri',
              calc,
              fxc='PBE',
              max_energy_diff=de,
              lr_communicators=lr_comms
              )
w, S, R, Sx, Sy, Sz = lr.get_transitions(max_energy=1e9, units='au')

e0_1 = w[0]
e1_1 = w[-1]
# Continue with larger de
de = 4.5
lr = LrTDDFT2('Al2_lri',
              calc,
              fxc='PBE',
              max_energy_diff=de,
              lr_communicators=lr_comms)
w, S, R, Sx, Sy, Sz = lr.get_transitions(max_energy=1e9, units='au')
e0_2 = w[0]
e1_2 = w[-1]
# Continue with smaller de
de = 2.5
lr = LrTDDFT2('Al2_lri',
              calc,
              fxc='PBE',
              max_energy_diff=de,
              lr_communicators=lr_comms)
w, S, R, Sx, Sy, Sz = lr.get_transitions(max_energy=1e9, units='au')
e0_3 = w[0]
e1_3 = w[-1]

if debug and rank == 0:
    print(e0_1, e1_1)
    print(e0_2, e1_2)
    print(e0_3, e1_3)

tol = 5.0e-8
equal(e0_1, e0_2, tol)
equal(e0_1, e0_3, tol)
tol = 1.0e-3
equal(e0_1, 0.00105074187176, tol)
equal(e1_1, 0.183188157301, tol)
equal(e1_2, 0.194973135812, tol)
equal(e1_3, 0.120681529342, tol)



from ase.build import molecule
from gpaw import GPAW
atoms = molecule('H2')
atoms.center(vacuum=2)
calc = GPAW(mode='lcao', basis='dzp', setups={0: 'paw', 1: 'ghost'}, xc='GLLBSC')
atoms.set_calculator(calc)
energy = atoms.get_potential_energy()



from ase.build import bulk

from gpaw import GPAW, FermiDirac
from gpaw.test import equal

a = 5.475
calc = GPAW(h=0.24,
            kpts=(4, 4, 4),
            poissonsolver={'name': 'fft'},
            occupations=FermiDirac(width=0.0),
            nbands=5)
atoms = bulk('Si', 'diamond', a=a)
atoms.set_calculator(calc)
E = atoms.get_potential_energy()
equal(E, -11.8699605591, 0.001)
niter = calc.get_number_of_iterations()

equal(atoms.calc.get_fermi_level(), 5.17751284, 0.005)
homo, lumo = calc.get_homo_lumo()
equal(lumo - homo, 1.11445025, 0.002)

calc.write('si_primitive.gpw', 'all')
calc = GPAW('si_primitive.gpw',
            parallel={'domain': 1, 'band': 1},
            idiotproof=False,
            txt=None)



# {   -1,   -1}:  On entry to PDLASRT parameter number    9 had an illegal value

# works with 'sl_default': (2, 2, 32)

from ase.build import fcc100, add_adsorbate
from gpaw import GPAW, ConvergenceError
from gpaw.mpi import world
from gpaw.utilities import compiled_with_sl

assert world.size == 4

slab = fcc100('Cu', size=(2, 2, 2))
add_adsorbate(slab,'O', 1.1, 'hollow')
slab.center(vacuum=3.0, axis=2)

calc = GPAW(mode='lcao',
            kpts=(2, 2, 1),
            txt='-',
            maxiter=1,)

if compiled_with_sl():
    calc.set(parallel={'domain': (1, 1, 4), 'sl_default': (2, 2, 64)})

slab.set_calculator(calc)
try:
    slab.get_potential_energy()
except ConvergenceError:
    pass



from __future__ import print_function, division

import numpy as np
from scipy.spatial import Delaunay

from ase.utils import devnull
from ase.utils import convert_string_to_fd
from ase.utils.timing import timer, Timer

from _gpaw import tetrahedron_weight

import gpaw.mpi as mpi
from gpaw.utilities.blas import gemm, rk, czher, mmm
from gpaw.utilities.progressbar import ProgressBar
from functools import partial


class Integrator():

    def __init__(self, cell_cv, response='density', comm=mpi.world,
                 txt='-', timer=None, nblocks=1, eshift=0.0):
        """Baseclass for Brillouin zone integration and band summation.

        Simple class to calculate integrals over Brilloun zones
        and summation of bands.

        comm: mpi.communicator
        nblocks: block parallelization
        """
        
        self.response = response
        self.comm = comm
        self.eshift = eshift
        self.nblocks = nblocks
        self.vol = abs(np.linalg.det(cell_cv))
        if nblocks == 1:
            self.blockcomm = self.comm.new_communicator([comm.rank])
            self.kncomm = comm
        else:
            assert comm.size % nblocks == 0, comm.size
            rank1 = comm.rank // nblocks * nblocks
            rank2 = rank1 + nblocks
            self.blockcomm = self.comm.new_communicator(range(rank1, rank2))
            ranks = range(comm.rank % nblocks, comm.size, nblocks)
            self.kncomm = self.comm.new_communicator(ranks)

        if comm.rank != 0:
            txt = devnull
        self.fd = convert_string_to_fd(txt, comm)

        self.timer = timer or Timer()

    def distribute_domain(self, domain_dl):
        """Distribute integration domain. """
        domainsize = [len(domain_l) for domain_l in domain_dl]
        nterms = np.prod(domainsize)
        size = self.kncomm.size
        rank = self.kncomm.rank

        n = (nterms + size - 1) // size
        i1 = rank * n
        i2 = min(i1 + n, nterms)
        mydomain = []
        for i in range(i1, i2):
            unravelled_d = np.unravel_index(i, domainsize)
            arguments = []
            for domain_l, index in zip(domain_dl, unravelled_d):
                arguments.append(domain_l[index])
            mydomain.append(tuple(arguments))

        print('Distributing domain %s' % (domainsize, ),
              'over %d process%s' %
              (self.kncomm.size, ['es', ''][self.kncomm.size == 1]),
              file=self.fd)
        print('Number of blocks:', self.blockcomm.size, file=self.fd)

        return mydomain

    def integrate(self, *args, **kwargs):
        raise NotImplementedError


class PointIntegrator(Integrator):

    def __init__(self, *args, **kwargs):

        """Integrate brillouin zone using a broadening technique.

        The broadening technique consists of smearing out the
        delta functions appearing in many integrals by some factor
        eta. In this code we use Lorentzians."""
        Integrator.__init__(self, *args, **kwargs)

    def integrate(self, kind='pointwise', *args, **kwargs):
        print('Integral kind:', kind, file=self.fd)
        if kind == 'pointwise':
            return self.pointwise_integration(*args, **kwargs)
        elif kind == 'hermitian response function':
            return self.response_function_integration(hermitian=True,
                                                      hilbert=False,
                                                      wings=False,
                                                      intraband=False,
                                                      *args, **kwargs)
        elif kind == 'hermitian response function wings':
            return self.response_function_integration(hermitian=True,
                                                      hilbert=False,
                                                      wings=True,
                                                      *args, **kwargs)
        elif kind == 'spectral function':
            return self.response_function_integration(hilbert=True,
                                                      *args, **kwargs)
        elif kind == 'spectral function wings':
            return self.response_function_integration(hilbert=True,
                                                      wings=True,
                                                      *args, **kwargs)
        elif kind == 'response function':
            return self.response_function_integration(hilbert=False,
                                                      *args, **kwargs)
        elif kind == 'response function wings':
            return self.response_function_integration(hilbert=False,
                                                      wings=True,
                                                      *args, **kwargs)
        else:
            raise NotImplementedError

    def response_function_integration(self, domain=None, integrand=None,
                                      x=None, kwargs=None, out_wxx=None,
                                      timeordered=False, hermitian=False,
                                      intraband=False, hilbert=False,
                                      wings=False, **extraargs):
        """Integrate a response function over bands and kpoints.

        func: method
        omega_w: ndarray
        out: np.ndarray
        timeordered: Bool
        """
        if out_wxx is None:
            raise NotImplementedError

        nG = out_wxx.shape[2]
        mynG = (nG + self.blockcomm.size - 1) // self.blockcomm.size
        self.Ga = min(self.blockcomm.rank * mynG, nG)
        self.Gb = min(self.Ga + mynG, nG)
        # assert mynG * (self.blockcomm.size - 1) < nG, \
        #     print('mynG', mynG, 'nG', nG, 'nblocks', self.blockcomm.size)

        mydomain_t = self.distribute_domain(domain)
        nbz = len(domain[0])
        get_matrix_element, get_eigenvalues = integrand
        
        prefactor = (2 * np.pi)**3 / self.vol / nbz
        out_wxx /= prefactor

        # The kwargs contain any constant
        # arguments provided by the user
        if kwargs is not None:
            get_matrix_element = partial(get_matrix_element,
                                         **kwargs[0])
            get_eigenvalues = partial(get_eigenvalues,
                                      **kwargs[1])

        # Sum kpoints
        # Calculate integrations weight
        pb = ProgressBar(self.fd)
        for _, arguments in pb.enumerate(mydomain_t):
            n_MG = get_matrix_element(*arguments)
            if n_MG is None:
                continue
            deps_M = get_eigenvalues(*arguments)

            if intraband:
                self.update_intraband(n_MG, out_wxx, **extraargs)
            elif hermitian and not wings:
                self.update_hermitian(n_MG, deps_M, x, out_wxx, **extraargs)
            elif hermitian and wings:
                self.update_optical_limit(n_MG, deps_M, x, out_wxx,
                                          **extraargs)
            elif hilbert and not wings:
                self.update_hilbert(n_MG, deps_M, x, out_wxx, **extraargs)
            elif hilbert and wings:
                self.update_hilbert_optical_limit(n_MG, deps_M, x,
                                                  out_wxx, **extraargs)
            elif wings:
                self.update_optical_limit(n_MG, deps_M, x, out_wxx,
                                          **extraargs)
            else:
                self.update(n_MG, deps_M, x, out_wxx, **extraargs)

        # Sum over
        self.kncomm.sum(out_wxx)

        if (hermitian or hilbert) and self.blockcomm.size == 1 and not wings:
            # Fill in upper/lower triangle also:
            nx = out_wxx.shape[1]
            il = np.tril_indices(nx, -1)
            iu = il[::-1]
            if hilbert:
                for out_xx in out_wxx:
                    out_xx[il] = out_xx[iu].conj()
            else:
                for out_xx in out_wxx:
                    out_xx[iu] = out_xx[il].conj()
        
        out_wxx *= prefactor

    @timer('CHI_0 update')
    def update(self, n_mG, deps_m, wd, chi0_wGG, timeordered=False, eta=None):
        """Update chi."""
        
        omega_w = wd.get_data()
        deps_m += self.eshift * np.sign(deps_m)
        if timeordered:
            deps1_m = deps_m + 1j * eta * np.sign(deps_m)
            deps2_m = deps1_m
        else:
            deps1_m = deps_m + 1j * eta
            deps2_m = deps_m - 1j * eta
        
        for omega, chi0_GG in zip(omega_w, chi0_wGG):
            if self.response == 'density':
                x_m = (1 / (omega + deps1_m) - 1 / (omega - deps2_m))
            else:
                x_m = - np.sign(deps_m) * 1. / (omega + deps1_m)
            if self.blockcomm.size > 1:
                nx_mG = n_mG[:, self.Ga:self.Gb] * x_m[:, np.newaxis]
            else:
                nx_mG = n_mG * x_m[:, np.newaxis]
             
            gemm(1.0, n_mG.conj(), np.ascontiguousarray(nx_mG.T),
                 1.0, chi0_GG)

    @timer('CHI_0 hermetian update')
    def update_hermitian(self, n_mG, deps_m, wd, chi0_wGG):
        """If eta=0 use hermitian update."""
        omega_w = wd.get_data()
        deps_m += self.eshift * np.sign(deps_m)

        for w, omega in enumerate(omega_w):
            if self.blockcomm.size == 1:
                x_m = (-2 * deps_m / (omega.imag**2 + deps_m**2) + 0j)**0.5
                nx_mG = n_mG.conj() * x_m[:, np.newaxis]
                rk(-1.0, nx_mG, 1.0, chi0_wGG[w], 'n')
            else:
                x_m = 2 * deps_m / (omega.imag**2 + deps_m**2)
                mynx_mG = n_mG[:, self.Ga:self.Gb] * x_m[:, np.newaxis]
                mmm(1.0, mynx_mG, 'C', n_mG, 'N', 1.0, chi0_wGG[w])

    @timer('CHI_0 spectral function update')
    def update_hilbert(self, n_mG, deps_m, wd, chi0_wGG):
        """Update spectral function.

        Updates spectral function A_wGG and saves it to chi0_wGG for
        later hilbert-transform."""

        self.timer.start('prep')
        omega_w = wd.get_data()
        deps_m += self.eshift * np.sign(deps_m)
        o_m = abs(deps_m)
        w_m = wd.get_closest_index(o_m)
        o1_m = omega_w[w_m]
        o2_m = omega_w[w_m + 1]
        p_m = np.abs(1 / (o2_m - o1_m)**2)
        p1_m = p_m * (o2_m - o_m)
        p2_m = p_m * (o_m - o1_m)
        self.timer.stop('prep')

        if self.blockcomm.size > 1:
            for p1, p2, n_G, w in zip(p1_m, p2_m, n_mG, w_m):
                if w + 1 < wd.wmax:  # The last frequency is not reliable
                    myn_G = n_G[self.Ga:self.Gb].reshape((-1, 1))
                    gemm(p1, n_G.reshape((-1, 1)), myn_G,
                         1.0, chi0_wGG[w], 'c')
                    gemm(p2, n_G.reshape((-1, 1)), myn_G,
                         1.0, chi0_wGG[w + 1], 'c')
            return

        for p1, p2, n_G, w in zip(p1_m, p2_m, n_mG, w_m):
            if w + 1 < wd.wmax:  # The last frequency is not reliable
                czher(p1, n_G.conj(), chi0_wGG[w])
                czher(p2, n_G.conj(), chi0_wGG[w + 1])

    @timer('CHI_0 intraband update')
    def update_intraband(self, vel_mv, chi0_wvv):
        """Add intraband contributions"""

        for vel_v in vel_mv:
            x_vv = np.outer(vel_v, vel_v)
            chi0_wvv[0] += x_vv

    @timer('CHI_0 optical limit update')
    def update_optical_limit(self, n_mG, deps_m, wd, chi0_wxvG,
                             timeordered=False, eta=None):
        """Optical limit update of chi."""

        omega_w = wd.get_data()
        if timeordered:
            # avoid getting a zero from np.sign():
            deps1_m = deps_m + 1j * eta * np.sign(deps_m + 1e-20)
            deps2_m = deps1_m
        else:
            deps1_m = deps_m + 1j * eta
            deps2_m = deps_m - 1j * eta

        for w, omega in enumerate(omega_w):
            x_m = (1 / (omega + deps1_m) - 1 / (omega - deps2_m))
            chi0_wxvG[w, 0] += np.dot(x_m * n_mG[:, :3].T, n_mG.conj())
            chi0_wxvG[w, 1] += np.dot(x_m * n_mG[:, :3].T.conj(), n_mG)

    @timer('CHI_0 optical limit hilbert-update')
    def update_hilbert_optical_limit(self, n_mG, deps_m, wd, chi0_wxvG):
        """Optical limit update of chi-head and -wings."""

        omega_w = wd.get_data()
        for deps, n_G in zip(deps_m, n_mG):
            o = abs(deps)
            w = wd.get_closest_index(o)
            if w + 2 > len(omega_w):
                break
            o1, o2 = omega_w[w:w + 2]
            if o > o2:
                continue
            else:
                assert o1 <= o <= o2, (o1, o, o2)

            p = 1 / (o2 - o1)**2
            p1 = p * (o2 - o)
            p2 = p * (o - o1)
            x_vG = np.outer(n_G[:3], n_G.conj())
            chi0_wxvG[w, 0, :, :] += p1 * x_vG
            chi0_wxvG[w + 1, 0, :, :] += p2 * x_vG
            chi0_wxvG[w, 1, :, :] += p1 * x_vG.conj()
            chi0_wxvG[w + 1, 1, :, :] += p2 * x_vG.conj()


class TetrahedronIntegrator(Integrator):
    """Integrate brillouin zone using tetrahedron integration.

    Tetrahedron integration uses linear interpolation of
    the eigenenergies and of the matrix elements
    between the vertices of the tetrahedron."""

    def __init__(self, *args, **kwargs):
        Integrator.__init__(self, *args, **kwargs)

    @timer('Tesselate')
    def tesselate(self, vertices):
        """Get tesselation descriptor."""
        td = Delaunay(vertices)

        td.volumes_s = None
        return td

    def get_simplex_volume(self, td, S):
        """Get volume of simplex S"""

        if td.volumes_s is not None:
            return td.volumes_s[S]

        td.volumes_s = np.zeros(td.nsimplex, float)
        for s in range(td.nsimplex):
            K_k = td.simplices[s]
            k_kc = td.points[K_k]
            volume = np.abs(np.linalg.det(k_kc[1:] - k_kc[0])) / 6.
            td.volumes_s[s] = volume

        return self.get_simplex_volume(td, S)

    def integrate(self, kind, *args, **kwargs):
        if kind == 'spectral function':
            return self.spectral_function_integration(*args, **kwargs)
        else:
            raise NotImplementedError

    @timer('Spectral function integration')
    def spectral_function_integration(self, domain=None, integrand=None,
                                      x=None, kwargs=None, out_wxx=None):
        """Integrate response function.

        Assume that the integral has the
        form of a response function. For the linear tetrahedron
        method it is possible calculate frequency dependent weights
        and do a point summation using these weights."""

        if out_wxx is None:
            raise NotImplementedError

        nG = out_wxx.shape[2]
        mynG = (nG + self.blockcomm.size - 1) // self.blockcomm.size
        self.Ga = min(self.blockcomm.rank * mynG, nG)
        self.Gb = min(self.Ga + mynG, nG)
        # assert mynG * (self.blockcomm.size - 1) < nG, \
        #     print('mynG', mynG, 'nG', nG, 'nblocks', self.blockcomm.size)

        # Input domain
        td = self.tesselate(domain[0])
        args = domain[1:]
        get_matrix_element, get_eigenvalues = integrand

        # The kwargs contain any constant
        # arguments provided by the user
        if kwargs is not None:
            get_matrix_element = partial(get_matrix_element,
                                         **kwargs[0])
            get_eigenvalues = partial(get_eigenvalues,
                                      **kwargs[1])

        # Relevant quantities
        bzk_kc = td.points
        nk = len(bzk_kc)

        with self.timer('pts'):
            # Point to simplex
            pts_k = [[] for n in range(nk)]
            for s, K_k in enumerate(td.simplices):
                A_kv = np.append(td.points[K_k],
                                 np.ones(4)[:, np.newaxis], axis=1)

                D_kv = np.append((A_kv[:, :-1]**2).sum(1)[:, np.newaxis],
                                 A_kv, axis=1)
                a = np.linalg.det(D_kv[:, np.arange(5) != 0])

                if np.abs(a) < 1e-10:
                    continue

                for K in K_k:
                    pts_k[K].append(s)

            # Change to numpy arrays:
            for k in range(nk):
                pts_k[k] = np.array(pts_k[k], int)

        with self.timer('neighbours'):
            # Nearest neighbours
            neighbours_k = [None for n in range(nk)]

            for k in range(nk):
                neighbours_k[k] = np.unique(td.simplices[pts_k[k]])

        # Distribute everything
        myterms_t = self.distribute_domain(list(args) +
                                           [list(range(nk))])

        with self.timer('eigenvalues'):
            # Store eigenvalues
            deps_tMk = None  # t for term
            shape = [len(domain_l) for domain_l in args]
            nterms = int(np.prod(shape))

            for t in range(nterms):
                if len(shape) == 0:
                    arguments = ()
                else:
                    arguments = np.unravel_index(t, shape)
                for K in range(nk):
                    k_c = bzk_kc[K]
                    deps_M = -get_eigenvalues(k_c, *arguments)
                    if deps_tMk is None:
                        deps_tMk = np.zeros([nterms] +
                                            list(deps_M.shape) +
                                            [nk], float)
                    deps_tMk[t, :, K] = deps_M

        omega_w = x.get_data()

        # Calculate integrations weight
        pb = ProgressBar(self.fd)
        for _, arguments in pb.enumerate(myterms_t):
            K = arguments[-1]
            if len(shape) == 0:
                t = 0
            else:
                t = np.ravel_multi_index(arguments[:-1], shape)
            deps_Mk = deps_tMk[t]
            teteps_Mk = deps_Mk[:, neighbours_k[K]]
            i0_M, i1_M = x.get_index_range(teteps_Mk.min(1), teteps_Mk.max(1))

            n_MG = get_matrix_element(bzk_kc[K],
                                      *arguments[:-1])
            for n_G, deps_k, i0, i1 in zip(n_MG, deps_Mk,
                                           i0_M, i1_M):
                if i0 == i1:
                    continue

                W_w = self.get_kpoint_weight(K, deps_k,
                                             pts_k, omega_w[i0:i1],
                                             td)

                for iw, weight in enumerate(W_w):
                    if self.blockcomm.size > 1:
                        myn_G = n_G[self.Ga:self.Gb].reshape((-1, 1))
                        gemm(weight, n_G.reshape((-1, 1)), myn_G,
                             1.0, out_wxx[i0 + iw], 'c')
                    else:
                        czher(weight, n_G.conj(), out_wxx[i0 + iw])

        self.kncomm.sum(out_wxx)

        if self.blockcomm.size == 1:
            # Fill in upper/lower triangle also:
            nx = out_wxx.shape[1]
            il = np.tril_indices(nx, -1)
            iu = il[::-1]
            for out_xx in out_wxx:
                out_xx[il] = out_xx[iu].conj()

    @timer('Get kpoint weight')
    def get_kpoint_weight(self, K, deps_k, pts_k,
                          omega_w, td):
        # Find appropriate index range
        simplices_s = pts_k[K]
        W_w = np.zeros(len(omega_w), float)
        vol_s = self.get_simplex_volume(td, simplices_s)
        with self.timer('Tetrahedron weight'):
            tetrahedron_weight(deps_k, td.simplices, K,
                               simplices_s,
                               W_w, omega_w, vol_s)

        return W_w
